#!/usr/bin/env python3
# coding=utf-8
#
# Copyright (C) 2005,2007,2008 Aaron Spike, aaron@ekips.org
# Copyright (C) 2008,2010 Alvin Penner, penner@vaxxine.com
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
#
"""
This file output script for Inkscape creates a AutoCAD R14 DXF file.
The spec can be found here: http://www.autodesk.com/techpubs/autocad/acadr14/dxf/index.htm.

 File history:
 - template dxf_outlines.dxf added Feb 2008 by Alvin Penner
- ROBO-Master output option added Aug 2008
- ROBO-Master multispline output added Sept 2008
- LWPOLYLINE output modification added Dec 2008
- toggle between LINE/LWPOLYLINE added Jan 2010
- support for transform elements added July 2010
- support for layers added July 2010
- support for rectangle added Dec 2010
"""

from __future__ import print_function

import inkex
from inkex import (
    colors,
    bezier,
    Transform,
    Group,
    Layer,
    Use,
    PathElement,
    Rectangle,
    Line,
    Circle,
    Ellipse,
)
from inkex.localization import inkex_gettext as _


def get_matrix(u, i, j):
    if j == i + 2:
        return (
            (u[i] - u[i - 1])
            * (u[i] - u[i - 1])
            / (u[i + 2] - u[i - 1])
            / (u[i + 1] - u[i - 1])
        )
    elif j == i + 1:
        return (
            (u[i] - u[i - 1]) * (u[i + 2] - u[i]) / (u[i + 2] - u[i - 1])
            + (u[i + 1] - u[i]) * (u[i] - u[i - 2]) / (u[i + 1] - u[i - 2])
        ) / (u[i + 1] - u[i - 1])
    elif j == i:
        return (
            (u[i + 1] - u[i])
            * (u[i + 1] - u[i])
            / (u[i + 1] - u[i - 2])
            / (u[i + 1] - u[i - 1])
        )
    else:
        return 0


def get_fit(u, csp, col):
    return (
        (1 - u) ** 3 * csp[0][col]
        + 3 * (1 - u) ** 2 * u * csp[1][col]
        + 3 * (1 - u) * u**2 * csp[2][col]
        + u**3 * csp[3][col]
    )


class DxfOutlines(inkex.OutputExtension):
    def add_arguments(self, pars):
        pars.add_argument("--tab")
        pars.add_argument("-R", "--ROBO", type=inkex.Boolean, default=False)
        pars.add_argument("-P", "--POLY", type=inkex.Boolean, default=False)
        pars.add_argument("-F", "--FLATTENBEZ", type=inkex.Boolean, default=False)
        pars.add_argument(
            "--unit_from_document", type=inkex.Boolean, default=True
        )  # px
        pars.add_argument("--units", default="px")  # px
        pars.add_argument("--encoding", dest="char_encode", default="latin_1")
        pars.add_argument("--layer_option", default="all")
        pars.add_argument("--layer_name")

        self.dxf = []
        self.handle = 255  # handle for DXF ENTITY
        self.layers = ["0"]
        self.layer = "0"  # mandatory layer
        self.layernames = []
        self.csp_old = [[0.0, 0.0]] * 4  # previous spline
        self.d = [0.0]  # knot vector
        self.poly = [[0.0, 0.0]]  # LWPOLYLINE data

    def dxf_add(self, str):
        self.dxf.append(str.encode(self.options.char_encode))

    def dxf_line(self, csp):
        """Draw a line in the DXF format"""
        self.handle += 1
        self.dxf_add(
            "  0\nLINE\n  5\n%x\n100\nAcDbEntity\n  8\n%s\n 62\n%d\n100\nAcDbLine\n"
            % (self.handle, self.layer, self.color)
        )
        self.dxf_add(
            " 10\n%f\n 20\n%f\n 30\n0.0\n 11\n%f\n 21\n%f\n 31\n0.0\n"
            % (csp[0][0], csp[0][1], csp[1][0], csp[1][1])
        )

    def LWPOLY_line(self, csp):
        if (
            abs(csp[0][0] - self.poly[-1][0]) > 0.0001
            or abs(csp[0][1] - self.poly[-1][1]) > 0.0001
            or self.color_LWPOLY != self.color
        ):  # THIS LINE IS NEW
            self.LWPOLY_output()  # terminate current polyline
            self.poly = [csp[0]]  # initiallize new polyline
            self.color_LWPOLY = self.color
            self.layer_LWPOLY = self.layer
        self.poly.append(csp[1])

    def LWPOLY_output(self):
        if len(self.poly) == 1:
            return
        self.handle += 1
        closed = 1
        if (
            abs(self.poly[0][0] - self.poly[-1][0]) > 0.0001
            or abs(self.poly[0][1] - self.poly[-1][1]) > 0.0001
        ):
            closed = 0
        self.dxf_add(
            "  0\nLWPOLYLINE\n  5\n%x\n100\nAcDbEntity\n  8\n%s\n 62\n%d\n100\nAcDbPolyline\n 90\n%d\n 70\n%d\n"
            % (
                self.handle,
                self.layer_LWPOLY,
                self.color_LWPOLY,
                len(self.poly) - closed,
                closed,
            )
        )
        for i in range(len(self.poly) - closed):
            self.dxf_add(
                " 10\n%f\n 20\n%f\n 30\n0.0\n" % (self.poly[i][0], self.poly[i][1])
            )

    def dxf_spline(self, csp):
        knots = 8
        ctrls = 4
        self.handle += 1
        self.dxf_add(
            "  0\nSPLINE\n  5\n%x\n100\nAcDbEntity\n  8\n%s\n 62\n%d\n100\nAcDbSpline\n"
            % (self.handle, self.layer, self.color)
        )
        self.dxf_add(" 70\n8\n 71\n3\n 72\n%d\n 73\n%d\n 74\n0\n" % (knots, ctrls))
        for i in range(2):
            for j in range(4):
                self.dxf_add(" 40\n%d\n" % i)
        for i in csp:
            self.dxf_add(" 10\n%f\n 20\n%f\n 30\n0.0\n" % (i[0], i[1]))

    def ROBO_spline(self, csp):
        """this spline has zero curvature at the endpoints, as in ROBO-Master"""
        if (
            abs(csp[0][0] - self.csp_old[3][0]) > 0.0001
            or abs(csp[0][1] - self.csp_old[3][1]) > 0.0001
            or abs(
                (csp[1][1] - csp[0][1]) * (self.csp_old[3][0] - self.csp_old[2][0])
                - (csp[1][0] - csp[0][0]) * (self.csp_old[3][1] - self.csp_old[2][1])
            )
            > 0.001
        ):
            self.ROBO_output()  # terminate current spline
            self.xfit = [csp[0][0]]  # initiallize new spline
            self.yfit = [csp[0][1]]
            self.d = [0.0]
            self.color_ROBO = self.color
            self.layer_ROBO = self.layer
        self.xfit += 3 * [0.0]
        self.yfit += 3 * [0.0]
        self.d += 3 * [0.0]
        for i in range(1, 4):
            j = len(self.d) + i - 4
            self.xfit[j] = get_fit(i / 3.0, csp, 0)
            self.yfit[j] = get_fit(i / 3.0, csp, 1)
            self.d[j] = self.d[j - 1] + bezier.pointdistance(
                (self.xfit[j - 1], self.yfit[j - 1]), (self.xfit[j], self.yfit[j])
            )
        self.csp_old = csp

    def ROBO_output(self):
        try:
            import numpy
            from numpy.linalg import solve
        except ImportError:
            inkex.errormsg(
                _(
                    "Failed to import the numpy or numpy.linalg modules. "
                    "These modules are required by the ROBO option."
                    "Please install them and try again."
                )
            )
            return

        if len(self.d) == 1:
            return
        fits = len(self.d)
        ctrls = fits + 2
        knots = ctrls + 4
        self.xfit += 2 * [0.0]  # pad with 2 endpoint constraints
        self.yfit += 2 * [0.0]
        self.d += 6 * [0.0]  # pad with 3 duplicates at each end
        self.d[fits + 2] = self.d[fits + 1] = self.d[fits] = self.d[fits - 1]

        solmatrix = numpy.zeros((ctrls, ctrls), dtype=float)
        for i in range(fits):
            solmatrix[i, i] = get_matrix(self.d, i, i)
            solmatrix[i, i + 1] = get_matrix(self.d, i, i + 1)
            solmatrix[i, i + 2] = get_matrix(self.d, i, i + 2)
        solmatrix[fits, 0] = self.d[2] / self.d[fits - 1]  # curvature at start = 0
        solmatrix[fits, 1] = -(self.d[1] + self.d[2]) / self.d[fits - 1]
        solmatrix[fits, 2] = self.d[1] / self.d[fits - 1]
        solmatrix[fits + 1, fits - 1] = (self.d[fits - 1] - self.d[fits - 2]) / self.d[
            fits - 1
        ]  # curvature at end = 0
        solmatrix[fits + 1, fits] = (
            self.d[fits - 3] + self.d[fits - 2] - 2 * self.d[fits - 1]
        ) / self.d[fits - 1]
        solmatrix[fits + 1, fits + 1] = (self.d[fits - 1] - self.d[fits - 3]) / self.d[
            fits - 1
        ]
        xctrl = solve(solmatrix, self.xfit)
        yctrl = solve(solmatrix, self.yfit)
        self.handle += 1
        self.dxf_add(
            "  0\nSPLINE\n  5\n%x\n100\nAcDbEntity\n  8\n%s\n 62\n%d\n100\nAcDbSpline\n"
            % (self.handle, self.layer_ROBO, self.color_ROBO)
        )
        self.dxf_add(
            " 70\n0\n 71\n3\n 72\n%d\n 73\n%d\n 74\n%d\n" % (knots, ctrls, fits)
        )
        for i in range(knots):
            self.dxf_add(" 40\n%f\n" % self.d[i - 3])
        for i in range(ctrls):
            self.dxf_add(" 10\n%f\n 20\n%f\n 30\n0.0\n" % (xctrl[i], yctrl[i]))
        for i in range(fits):
            self.dxf_add(" 11\n%f\n 21\n%f\n 31\n0.0\n" % (self.xfit[i], self.yfit[i]))

    def process_shape(self, node, mat):
        if not isinstance(node, (PathElement, Rectangle, Line, Circle, Ellipse)):
            return

        rgb = (0, 0, 0)
        style = node.style("stroke")
        if style is not None and isinstance(style, inkex.Color):
            rgb = style.to_rgb()
        hsl = colors.rgb_to_hsl(rgb[0] / 255.0, rgb[1] / 255.0, rgb[2] / 255.0)
        self.color = 7  # default is blac
        if hsl[2]:
            self.color = 1 + (int(6 * hsl[0] + 0.5) % 6)  # use 6 hues

        # Transforming /after/ superpath is more reliable than before
        # because of some issues with arcs in transformations
        path = node.path.to_superpath().transform(Transform(mat) @ node.transform)

        # If Flatten Beziers is enabled, subdivide our beziers and
        # we'll later just ignore the curve and output flat lines
        if self.options.FLATTENBEZ:
            bezier.cspsubdiv(path, 0.1)  # default to most detailed (0.1)

        # Now output the path.
        for sub in path:
            for i in range(len(sub) - 1):
                s = sub[i]
                e = sub[i + 1]
                # If flattening beziers, ignore curves and output flat lines
                if (s[1] == s[2] and e[0] == e[1]) or self.options.FLATTENBEZ:
                    if self.options.POLY:
                        self.LWPOLY_line([s[1], e[1]])
                    else:
                        self.dxf_line([s[1], e[1]])
                elif self.options.ROBO:
                    self.ROBO_spline([s[1], s[2], e[0], e[1]])
                else:
                    self.dxf_spline([s[1], s[2], e[0], e[1]])

    def process_clone(self, node):
        """Process a clone node, looking for internal paths"""
        trans = node.get("transform")
        x = node.get("x")
        y = node.get("y")
        mat = Transform([[1.0, 0.0, 0.0], [0.0, 1.0, 0.0]])
        if trans:
            mat @= Transform(trans)
        if x:
            mat @= Transform([[1.0, 0.0, float(x)], [0.0, 1.0, 0.0]])
        if y:
            mat @= Transform([[1.0, 0.0, 0.0], [0.0, 1.0, float(y)]])
        # push transform
        if trans or x or y:
            self.groupmat.append(Transform(self.groupmat[-1]) @ mat)
        # get referenced node
        refid = node.get("xlink:href")
        refnode = self.svg.getElementById(refid[1:])
        if refnode is not None:
            if isinstance(refnode, Group):
                self.process_group(refnode)
            elif isinstance(refnode, Use):
                self.process_clone(refnode)
            else:
                self.process_shape(refnode, self.groupmat[-1])
        # pop transform
        if trans or x or y:
            self.groupmat.pop()

    def process_group(self, group):
        """Process group elements"""
        if isinstance(group, Layer):
            style = group.style
            if (
                style("display") == "none"
                and self.options.layer_option
                and self.options.layer_option == "visible"
            ):
                return
            layer = group.label
            if self.options.layer_name and self.options.layer_option == "name":
                if not layer.lower() in self.options.layer_name:
                    return

            layer = layer.replace(" ", "_")
            if layer in self.layers:
                self.layer = layer
        trans = group.get("transform")
        if trans:
            self.groupmat.append(Transform(self.groupmat[-1]) @ Transform(trans))
        for node in group:
            try:
                if isinstance(node, Group):
                    self.process_group(node)
                elif isinstance(node, Use):
                    self.process_clone(node)
                else:
                    self.process_shape(node, self.groupmat[-1])
            except RecursionError as e:
                raise inkex.AbortExtension(
                    _(
                        'Too many nested groups. Please use the "Deep Ungroup" extension first.'
                    )
                ) from e  # pylint: disable=line-too-long
        if trans:
            self.groupmat.pop()

    def save(self, stream):
        # Warn user if name match field is empty
        if (
            self.options.layer_option
            and self.options.layer_option == "name"
            and not self.options.layer_name
        ):
            return inkex.errormsg(
                _(
                    "Error: Field 'Layer match name' must be filled when using "
                    "'By name match' option"
                )
            )

        if len(self.svg.xpath("//svg:use|//svg:flowRoot|//svg:text")) > 0:
            self.preprocess(["flowRoot", "text"])

        # Split user layer data into a list: "layerA,layerb,LAYERC" becomes ["layera", "layerb", "layerc"]
        if self.options.layer_name:
            self.options.layer_name = self.options.layer_name.lower().split(",")

        # References:   Minimum Requirements for Creating a DXF File of a 3D Model By Paul Bourke
        #              NURB Curves: A Guide for the Uninitiated By Philip J. Schneider
        #              The NURBS Book By Les Piegl and Wayne Tiller (Springer, 1995)
        # self.dxf_add("999\nDXF created by Inkscape\n")  # Some programs do not take comments in DXF files (KLayout 0.21.12 for example)
        if self.options.unit_from_document:
            unit = self.svg.document_unit
        else:
            unit = self.options.units
        with open(self.get_resource("dxf14_header.txt"), "r") as fhl:
            header = fhl.read()
            unit_map = {"px": 0, "in": 1, "ft": 2, "mm": 4, "cm": 5, "m": 6}
            header = header.replace("<unit specifier>", str(unit_map.get(unit, 0)))
            self.dxf_add(header)
        for node in self.svg.xpath("//svg:g"):
            if isinstance(node, Layer):
                layer = node.label
                self.layernames.append(layer.lower())
                if (
                    self.options.layer_name
                    and self.options.layer_option
                    and self.options.layer_option == "name"
                    and not layer.lower() in self.options.layer_name
                ):
                    continue
                layer = layer.replace(" ", "_")
                if layer and layer not in self.layers:
                    self.layers.append(layer)
        self.dxf_add(
            "  2\nLAYER\n  5\n2\n100\nAcDbSymbolTable\n 70\n%s\n" % len(self.layers)
        )
        for i in range(len(self.layers)):
            self.dxf_add(
                "  0\nLAYER\n  5\n%x\n100\nAcDbSymbolTableRecord\n100\nAcDbLayerTableRecord\n  2\n%s\n 70\n0\n  6\nCONTINUOUS\n"
                % (i + 80, self.layers[i])
            )
        with open(self.get_resource("dxf14_style.txt"), "r") as fhl:
            self.dxf_add(fhl.read())

        # Set toplevel transform
        scale = self.svg.inkscape_scale
        self.groupmat = [
            [[scale, 0.0, 0.0], [0.0, -scale, self.svg.viewbox_height * scale]]
        ]
        self.process_group(self.svg)
        if self.options.ROBO:
            self.ROBO_output()
        if self.options.POLY:
            self.LWPOLY_output()
        with open(self.get_resource("dxf14_footer.txt"), "r") as fhl:
            self.dxf_add(fhl.read())
        # Warn user if layer data seems wrong
        if (
            self.options.layer_name
            and self.options.layer_option
            and self.options.layer_option == "name"
        ):
            for layer in self.options.layer_name:
                if layer not in self.layernames:
                    inkex.errormsg(_("Warning: Layer '{}' not found!").format(layer))

        stream.write(b"".join(self.dxf))


if __name__ == "__main__":
    DxfOutlines().run()



"""
VI decoders

converts Linearity Curve (5.x) JSON data to inkvn.
This handles Curve documents with flattened JSON (IDs).

Needs files that are fileFormatVersion 30~39.
"""

import base64
import logging
from typing import Any, Dict, List, Optional, Union

import inkex

import inkvn.reader.extract as ext
import inkvn.reader.text as t
from inkvn.utils import NSKeyedUnarchiver

from ..elements.artboard import VNArtboard, VNLayer, Frame
from ..elements.base import VNBaseElement, VNTransform
from ..elements.guide import VNGuideElement
from ..elements.group import VNGroupElement
from ..elements.image import VNImageElement
from ..elements.path import VNPathElement, pathGeometry
from ..elements.text import VNTextElement, singleStyledText, textProperty
from ..elements.styles import VNColor, VNGradient, pathStrokeStyle, basicStrokeStyle


def get_json_element(json_data: Dict, list_key: str, index: str) -> Optional[Dict]:
    """
    Retrieve an attribute according to key and index from gid_json.

    Linearity Curve makes use of IDs and lists of elements.
    This function makes it easier to get Curve elements.
    """

    elements = json_data.get(list_key)

    if elements is None:  # return None if the top-level key doesn't exist.
        return None

    return elements[index]


def read_artboard(archive: Any, gid_json: Dict) -> VNArtboard:
    """
    Reads gid.json(artboard) and returns artboard class.

    Argument `archive` is needed for image embedding.
    """
    # Layers
    # "layer_ids" contain layer indexes, while "layers" contain existing layers
    artboard = gid_json["artboards"][0]
    layer_ids = artboard["layerIds"]
    existing_layers = gid_json["layers"]
    layer_list: List[VNLayer] = []
    for layer_id in layer_ids:
        layer = get_json_element(gid_json, "layers", layer_id)
        if layer is not None:
            layer_list.append(read_layer(archive, gid_json, layer))

    # Guides
    # the very last layer has the guide
    guide_layer = existing_layers[len(layer_list)]
    guide_ids = guide_layer["elementIds"]
    guide_list: List[VNGuideElement] = []
    for guide_id in guide_ids:
        guide = get_json_element(gid_json, "elements", guide_id)
        if guide is not None:
            guide_list.append(read_element(archive, gid_json, guide))

    # Background Color
    fill_color = None
    fill_gradient = None
    fill_id = artboard.get("fillId")
    if fill_id is not None:
        fill_result = read_fill(gid_json, artboard, fill_id)
        if isinstance(fill_result, VNGradient):
            fill_gradient = fill_result
        elif isinstance(fill_result, VNColor):
            fill_color = fill_result

    return VNArtboard(
        title=artboard["title"],
        frame=Frame(**artboard["frame"]),
        layers=layer_list,
        guides=guide_list,
        fillColor=fill_color,
        fillGradient=fill_gradient,
    )


def read_layer(archive: Any, gid_json: Dict, layer: Dict) -> VNLayer:
    """
    Read specified layer and return their attributes as class.

    gid_json is used for finding elements inside the layer.
    """
    layer_element_ids = layer["elementIds"]
    element_list: List[VNBaseElement] = []
    for element_id in layer_element_ids:
        element = get_json_element(gid_json, "elements", element_id)
        if element is not None:
            element_list.append(read_element(archive, gid_json, element))

    return VNLayer(
        name=layer.get("name", "Unnamed Layer"),
        opacity=layer.get("opacity", 1.0),
        isVisible=layer.get("isVisible", True),
        isLocked=layer.get("isLocked", False),
        isExpanded=layer.get("isExpanded", False),
        elements=element_list,
    )


def read_element(archive: Any, gid_json: Dict, element: Dict) -> VNBaseElement:
    """Traverse specified element and extract their attributes."""
    base_element_data = {
        "name": element.get("name", "Unnamed Element"),
        "blur": element.get("blur", 0.0),
        "opacity": element.get("opacity", 1.0),
        "blendMode": element.get("blendMode", 0),
        "isHidden": element.get("isHidden", False),
        "isLocked": element.get("isLocked", False),
        "localTransform": None,
    }

    try:
        # localTransform (BaseElement)
        local_transform_id = element["localTransformId"]
        if local_transform_id is not None:
            base_element_data["localTransform"] = VNTransform(
                **get_json_element(gid_json, "localTransforms", local_transform_id)
            )

        # Guide (GuideElement)
        guide_id = element.get("subElement", {}).get("guideLine", {}).get("_0")
        if guide_id is not None:
            guide = get_json_element(gid_json, "guideLines", guide_id)
            return VNGuideElement(**guide, **base_element_data)

        # Group (GroupElement)
        group_id = element.get("subElement", {}).get("group", {}).get("_0")
        if group_id is not None:
            return read_group(archive, gid_json, group_id, base_element_data)

        # Image (ImageElement)
        image_id = element.get("subElement", {}).get("image", {}).get("_0")
        if image_id is not None:
            return read_image(archive, gid_json, image_id, base_element_data)

        # abstractImage in Curve 5.13.0, format 40
        abst_image_id = element.get("subElement", {}).get("abstractImage", {}).get("_0")
        if abst_image_id is not None:
            return read_image(archive, gid_json, abst_image_id, base_element_data)

        # Stylable (either PathElement or TextElement)
        stylable_id = element.get("subElement", {}).get("stylable", {}).get("_0")
        if stylable_id is not None:
            stylable = get_json_element(gid_json, "stylables", stylable_id)

            # clipping mask
            mask = stylable.get("mask", 0)

            # will be used to return PathElement
            fill_id = None
            stroke_style_id = None
            abstract_path_id = None

            # singleStyles (based on Curve 5.1.2)
            single_style_id = (
                stylable.get("subElement", {}).get("singleStyle", {}).get("_0")
            )
            if single_style_id is not None:
                single_style = get_json_element(
                    gid_json, "singleStyles", single_style_id
                )
                # old abstractPath lacks these ids below
                abstract_path_id = single_style.get("subElement")
                stroke_style_id = stylable.get("strokeStyleId")
                fill_id = single_style.get("fillId")

            # Abstract Path (PathElement)
            sub_element = stylable.get("subElement", {})
            if "abstractPath" in sub_element:
                abstract_path_id = sub_element["abstractPath"].get("_0")
            if abstract_path_id is not None:
                path_element_data = {
                    "abst_path": abstract_path_id,
                    "stylable": stylable,
                    "mask": mask,
                    "stroke": stroke_style_id,
                    "fill": fill_id,
                }
                return read_abst_path(gid_json, path_element_data, base_element_data)

            # Abstract Text (TextElement)
            abstract_text_id = (
                stylable.get("subElement", {}).get("abstractText", {}).get("_0")
            )
            if abstract_text_id is not None:
                return read_abst_text(gid_json, abstract_text_id, base_element_data)

        # if the element is unknown type:
        raise NotImplementedError(
            f"{base_element_data['name']}: This element has unknown type."
        )

    except Exception as e:
        logging.error(f"Error reading element: {e}", exc_info=True)
        return VNBaseElement(**base_element_data)


def read_group(
    archive: Any, gid_json: Dict, group_id: int, base_element: Dict
) -> VNGroupElement:
    """Reads elements inside group and returns as VNGroupElement."""
    # get elements inside group
    group = get_json_element(gid_json, "groups", group_id)
    group_element_ids = group["elementIds"]
    group_element_list: List[VNBaseElement] = []

    for group_element_id in group_element_ids:
        group_element = get_json_element(gid_json, "elements", group_element_id)
        if group_element is not None:
            # get group elements recursively
            group_element_list.append(read_element(archive, gid_json, group_element))

    return VNGroupElement(groupElements=group_element_list, **base_element)


def read_image(
    archive: Any, gid_json: Dict, image_id: int, base_element: Dict
) -> VNImageElement:
    """Reads image element, encodes image in Base64 and returns VNImageElement."""
    # relativePath contains *.dat (bitmap data)
    # sharedFileImage doesn't exist in 5.1.1 (file version 21) document
    # Curve 5.13.0, format 40 uses abstractImage
    # TODO format 40 support lags behind other versions(14, 19, 44)
    transform = None
    image_data_id = None
    crop_rect = None

    image = get_json_element(gid_json, "images", image_id)
    abst_image = get_json_element(gid_json, "abstractImages", image_id)

    if image is not None:
        if image.get("imageData") is not None:
            image_data_id = (
                image.get("imageData", {}).get("sharedFileImage", {}).get("_0")
            )
            # cropping
            if image.get("cropRect") is not None:
                crop_rect = tuple(map(tuple, image.get("cropRect")))
        else:  # legacy image
            image_data_id = image.get("imageDataId")
            transform = image.get("transform")

    elif abst_image is not None:
        image_data_id = abst_image.get("subElement", {}).get("image", {}).get("_0")
        transform = abst_image.get("transform")

        # cropping
        if abst_image.get("cropRect") is not None:
            crop_rect = tuple(map(tuple, abst_image.get("cropRect")))

    image_file = get_json_element(gid_json, "imageDatas", image_data_id)["relativePath"]
    image_data = ext.read_dat_from_zip(archive, image_file)
    return VNImageElement(
        imageData=image_data, transform=transform, cropRect=crop_rect, **base_element
    )


def read_abst_path(
    gid_json: Dict, path_element: Dict, base_element: Dict
) -> VNPathElement:
    """Reads path element and returns VNPathElement."""
    abstract_path = get_json_element(
        gid_json, "abstractPaths", path_element["abst_path"]
    )
    stroke_id = path_element["stroke"]
    fill_id = path_element["fill"]
    fill_color = None
    fill_gradient = None
    stroke_style = None

    # Stroke Style
    if "strokeStyleId" in abstract_path:
        stroke_id = abstract_path["strokeStyleId"]
    if stroke_id is not None:
        stroke_style = read_stroke(gid_json, stroke_id)

    # fill
    if "fillId" in abstract_path:
        fill_id = abstract_path["fillId"]
    if fill_id is not None:
        fill_result = read_fill(gid_json, path_element["stylable"], fill_id)
        if isinstance(fill_result, VNGradient):
            fill_gradient = fill_result
        elif isinstance(fill_result, VNColor):
            fill_color = fill_result

    # Path
    path_id = abstract_path.get("subElement", {}).get("path", {}).get("_0")
    path_geometry_list: List[pathGeometry] = []
    if path_id is not None:
        path = get_json_element(gid_json, "paths", path_id)

        # Path Geometry
        geometry_id = path.get("geometryId")
        if geometry_id is not None:
            path_geometry_list.append(
                pathGeometry(
                    **get_json_element(gid_json, "pathGeometries", geometry_id)
                )
            )

    # compoundPath
    compound_path_id = (
        abstract_path.get("subElement", {}).get("compoundPath", {}).get("_0")
    )
    if compound_path_id is not None:
        compound_path = get_json_element(gid_json, "compoundPaths", compound_path_id)

        # Path Geometries (subpath)
        subpath_ids = compound_path.get("subpathIds")
        if subpath_ids is not None:
            for id in subpath_ids:
                path_geometry_list.append(
                    pathGeometry(**get_json_element(gid_json, "pathGeometries", id))
                )

    return VNPathElement(
        mask=path_element["mask"],
        fillColor=fill_color,
        fillGradient=fill_gradient,
        strokeStyle=stroke_style,
        pathGeometries=path_geometry_list,
        **base_element,
    )


def read_abst_text(
    gid_json: Dict, abstract_text_id: int, base_element: Dict
) -> Union[VNTextElement, VNBaseElement]:
    """
    Reads Curve text element and returns VNTextElement.
    """
    # TODO Improve Text support, new format
    abstract_text = get_json_element(gid_json, "abstractTexts", abstract_text_id)
    # check if the text is new format
    if abstract_text.get("attributedText") is None:
        # Which one is which?
        styled_text_id = abstract_text.get("textId")
        text_id = abstract_text.get("subElement", {}).get("text", {}).get("_0")
        stroke_style_id = abstract_text.get("strokeStyleId")

        # will be used to return TextElement
        text_property = None
        styled_text = None

        # textPath
        text_path_id = abstract_text.get("subElement", {}).get("textPath", {}).get("_0")
        if text_path_id is not None:
            inkex.utils.debug(f"{base_element['name']}: textOnPath is not supported.")

        # texts(layout??), will be named textProperty internally
        if text_id is not None:
            text_property = get_json_element(gid_json, "texts", text_id)
            text_property = textProperty(
                textFrameLimits=text_property.get("textFrameLimits"),
                textFramePivot=text_property.get("textFramePivot"),
            )

        # text stroke_style only contains basicStrokeStyle
        if stroke_style_id is not None:
            stroke_style = get_json_element(
                gid_json, "textStrokeStyles", stroke_style_id
            )
            stroke_style = basicStrokeStyle(**stroke_style)

        # styledTexts
        if styled_text_id is not None:
            styled_text = get_json_element(gid_json, "styledTexts", styled_text_id)
            string = styled_text["string"]
            styles = t.decode_new_text(styled_text)
            styled_text_list = read_styled_text(styles)

        return VNTextElement(
            string=string,
            transform=None,
            styledText=styled_text_list,
            textProperty=text_property,
            **base_element,
        )

    # legacy text
    # TODO implement Legacy Text decoder
    else:
        # will be used to return TextElement
        transform = None
        text_property = None
        styled_text = None

        # I cannot replicate textProperty in legacy format
        text_id = abstract_text.get("subElement", {}).get("text", {}).get("_0")
        if text_id is not None:
            text_property = get_json_element(gid_json, "texts", text_id)
            transform = text_property.get("transform")  # matrix
            # resize_mode = text_property.get("resizeMode")
            # height = text_property.get("height")
            # width = text_property.get("width")

        # styledText
        styled_text = NSKeyedUnarchiver(
            base64.b64decode(abstract_text["attributedText"])
        )
        string = styled_text["NSString"]
        styles = t.decode_old_text(styled_text)
        styled_text_list = read_styled_text(styles)

        return VNTextElement(
            string=string,
            transform=transform,
            styledText=styled_text_list,
            textProperty=text_property,  # TODO Illegal format
            **base_element,
        )


def read_styled_text(styles: List[Dict]) -> List[singleStyledText]:
    styled_text_list: List[singleStyledText] = []
    for style in styles:
        color = None
        # color
        if style.get("fillColor") is not None:
            color = VNColor(style["fillColor"])
        # stroke # TODO text strokestyle(fix reader/text.py)
        # if style.get("strokeStyle") is not None:
        #    stroke = style["strokeStyle"]
        #    stroke = pathStrokeStyle(stroke_style, stroke)

        styled_text = singleStyledText(
            length=style.get("length"),
            fontName=style.get("fontName"),
            fontSize=style.get("fontSize"),
            alignment=style.get("alignment"),
            kerning=style.get("kerning"),
            lineHeight=style.get("lineHeight"),
            fillColor=color,
            fillGradient=None,  # TODO gradient applies globally
            strokeStyle=None,
            strikethrough=style.get("strikethrough"),
            underline=style.get("underline"),
        )
        styled_text_list.append(styled_text)

    return styled_text_list


def read_stroke(gid_json: Dict, stroke_id: int) -> pathStrokeStyle:
    """Reads stroke style and returns as class."""
    stroke_style = get_json_element(gid_json, "pathStrokeStyles", stroke_id)

    if stroke_style is None:  # Try legacy "strokeStyles" if not found
        stroke_style = get_json_element(gid_json, "strokeStyles", stroke_id)

    if (
        "dashPattern" in stroke_style
        and "join" in stroke_style
        and "cap" in stroke_style
    ):
        stroke_style["basicStrokeStyle"] = {
            "cap": stroke_style["cap"],
            "dashPattern": stroke_style["dashPattern"],
            "join": stroke_style["join"],
            "position": stroke_style["position"],
        }
    return pathStrokeStyle(
        basicStrokeStyle=basicStrokeStyle(**stroke_style["basicStrokeStyle"]),
        color=VNColor(color_dict=stroke_style["color"]),
        width=stroke_style["width"],
    )


def read_fill(
    gid_json: Dict, stylable: Dict, fill_id: int
) -> Union[VNGradient, VNColor, None]:
    """Reads fill data and returns as class."""
    color: Dict = (
        get_json_element(gid_json, "fills", fill_id).get("color", {}).get("_0")
    )
    gradient: Dict = (
        get_json_element(gid_json, "fills", fill_id).get("gradient", {}).get("_0")
    )

    if gradient is not None:
        # Newer Curve
        if gradient.get("transform") is not None:
            return VNGradient(
                fill_transform=gradient["transform"],
                transform_matrix=None,
                stops=gradient["gradient"]["stops"],
                typeRawValue=gradient["gradient"]["typeRawValue"],
            )
        # Old Curve like 5.1.1
        else:
            fill_transform_id = stylable.get("fillTransformId")
            fill_transform = get_json_element(
                gid_json, "fillTransforms", fill_transform_id
            )
            return VNGradient(
                fill_transform=fill_transform,
                transform_matrix=fill_transform["transform"],
                stops=gradient["stops"],
                typeRawValue=gradient["typeRawValue"],
            )
    elif color is not None:
        return VNColor(color_dict=color)



#!/usr/bin/env python3

# Copyright (C) 2024 Jonathan Neuhauser <jonathan.neuhauser@outlook.com>
# SPDX-License-Identifier: GPL-2.0-or-later

"""
Reads an afdesign file and extracts the contents, to be used by the
parser.
"""

from __future__ import annotations
from io import BufferedReader
from typing import Dict, List, Optional
from dataclasses import dataclass
from datetime import datetime
from enum import Enum
import warnings
import zlib
import zstandard


from .utils import BinaryParser
from .consts import expected_file_classes, expected_documents, file_magic


@dataclass
class InfHeader:
    """Structure to handle basic data in the Header"""

    fat_offset: int
    thumbnail_offset: int
    creation_date: datetime
    revision: int


@dataclass
class InfHeader2:
    """Structure to handle additional data in the Header"""

    revision: int


@dataclass
class Header:
    """Header of an .afdesign file"""

    version: int
    class_tag: str
    inf_header: InfHeader
    inf_header_2: Optional[InfHeader2]


@dataclass
class AFFileStructure:
    """Struct for an .afdesign file"""

    header: Header
    files: Dict[str, bytes]


@dataclass
class FAT:
    """Struct for the file access table (FAT)"""

    tag: str
    creation_date: datetime
    files: List[FATFile]
    dirs: List[FATDir]

    def get_file(self, file_id: int) -> Optional[FATFile]:
        """Finds the file with the given id, or None"""
        return next((file for file in self.files if file.id == file_id), None)

    def __repr__(self):
        return f"{{Date: {str(self.creation_date)}, Files: {len(self.files)}, Directories: {len(self.dirs)}}}"


class FATFileType(Enum):
    """Type of a FAT file"""

    NORMAL_WITH_NAME = 0
    NORMAL_WITHOUT_NAME = 1
    DELETED = 2


@dataclass
class FATFile:
    """Struct for an individual file. This class is directly
    used only for files with Type == Deleted"""

    id: int
    flag: FATFileType


@dataclass
class NormalFATFile(FATFile):
    """Struct for an individual file"""

    offset: int
    size: int
    compressed_size: int
    crc32: int
    compression: int
    name: Optional[str] = None


@dataclass
class FATDir:
    """Struct for a directory within the FAT"""

    files_num: int
    name: str


def undo_cumulative_sum(data: bytes) -> bytes:
    byte = 0
    result = bytearray()
    for b in data:
        byte = (byte + b) & 0xFF
        result.append(byte)
    return bytes(result)


class AFExtractor(BinaryParser):
    """Extract the contents of the .afdesign archive.
    Afterwards, self.content.files contains the binary
    data for the latest revision of all files in the archive"""

    def __init__(self, data: BufferedReader) -> None:
        super().__init__(data)
        self.content: Optional[AFFileStructure] = None
        self.ids: Dict[str, int] = {}
        self.revisions: List[datetime] = []
        self.fats: List[FAT] = []
        self._read_file()

    def _read_file(self):
        self.data.stream.seek(0)
        header = self._read_header()
        self.content = AFFileStructure(header, {})

        self._load_fat()
        self._load_content()

    def _read_header(self) -> Header:
        assert self.read_u32() == file_magic
        version = self.read_u16()
        flag = self.read_u16()
        class_tag = self.read_tag()
        assert class_tag in expected_file_classes, f"unknown file type {class_tag}"
        assert 7 <= version <= 11, "invalid version"

        # read the header information
        header_tag = self.read_tag()
        assert header_tag == "fnI#", f"Unexpected header tag {header_tag}"
        fat_offset = self.read_u64()
        thmb_offset = self.read_u64()
        self.data.read(16)  # unknown
        timestamp = self.read_u64()
        revision = self.read_u32()
        self.read_u32()  # unknown

        inf2 = None
        if version > 7:
            assert self.read_tag() == "torP"
            revision2 = self.read_u32()
            inf2 = InfHeader2(revision2)

        result = Header(
            version,
            class_tag,
            InfHeader(
                fat_offset, thmb_offset, datetime.fromtimestamp(timestamp), revision
            ),
            inf2,
        )

        if not (flag % 2 == 0 and flag & 2 == 0):
            raise ValueError(f"Unknown format, flag was {flag}")

        return result

    def _load_fat_file(self, fat_tag: str):
        fat_id = self.read_u32()
        flag = FATFileType(self.read_u8())
        if flag != FATFileType.DELETED:
            file_offset = self.read_u64()
            size = self.read_u64()
            compressed_size = self.read_u64()
            crc32 = self.read_u32()
            compression = self.read_u8()
            if not fat_tag == "TAF#":
                self.read(4)  # Unknown u32
            if fat_tag == "4TF#":
                self.read(4)  # Unknown u32
            if fat_tag in ("TAF#", "2TF#"):
                compression = {
                    0x01: 0x01,
                    0x02: 0x41,
                    0x03: 0x81,
                    0x04: 0xC1,
                }.get(compression, 0)
            name = None
            if flag == FATFileType.NORMAL_WITH_NAME:
                name = self.read(self.read_u16()).decode("latin-1")
                if name in self.ids:
                    assert self.ids[name] == fat_id, "Invalid ID"
                self.ids[name] = fat_id
            return NormalFATFile(
                fat_id,
                flag,
                file_offset,
                size,
                compressed_size,
                crc32,
                compression,
                name,
            )
        return FATFile(fat_id, flag)

    def _load_fat(self):
        """Load the file access table"""
        assert self.content is not None
        offset = self.content.header.inf_header.fat_offset
        while offset != 0:
            self.data.stream.seek(offset)
            fat_tag = self.read_tag()
            assert fat_tag in (
                "TAF#",
                "2TF#",
                "3TF#",
                "4TF#",
            ), f"unexpected FAT tag: {fat_tag}"
            next_offset = self.read_u64()
            created = datetime.fromtimestamp(self.read_u64())
            self.read(24)  # Unknown, 3x u64
            files_count = self.read_u32()
            self.read(8)  # Unknown, 2x u32
            dirs_count = self.read_u16()
            self.read(1)  # Unknown, 1x u8

            fat_files = []
            fat_dirs = []

            for _ in range(files_count):
                fat_files.append(self._load_fat_file(fat_tag))

            for _ in range(dirs_count):
                str_len = self.read_u16()
                some_len = self.read_u16()
                assert some_len == 0
                files_num = self.read_u64()
                name = self.read(str_len).decode("latin-1")
                fat_dirs.append(FATDir(files_num, name))

            self.revisions.append(created)
            self.fats.append(FAT(fat_tag, created, fat_files, fat_dirs))
            offset = next_offset

    def get_head_revision(self, name):
        return self._get_revision(name, None)

    def _get_revision(
        self, name, revision_date: Optional[datetime]
    ) -> Optional[FATFile]:
        if name not in self.ids:
            return None
        fat_id = self.ids[name]
        # Find the FAT where the latest revision before revision_date of file sits.
        # by sorting descending by creation date
        desc = sorted(
            (
                (fat.creation_date, fat.get_file(fat_id))
                for fat in self.fats
                if fat.get_file(fat_id) is not None
                and (revision_date is None or fat.creation_date <= revision_date)
            ),
            key=lambda i: i[0],
            reverse=True,
        )
        if not desc:
            return None
        return desc[0][1]

    def _load_content(self):
        for doc in expected_documents:
            revision = self.get_head_revision(doc)
            if revision is None:
                continue
            if not isinstance(revision, NormalFATFile):
                continue
            if revision.compressed_size == 0:
                continue

            file_content = self.extract(revision)

            assert self.content is not None
            self.content.files[doc] = file_content

    def extract(self, file: NormalFATFile) -> bytes:
        """Extract a FAT file"""
        assert file.size != 0, "Empty file"
        assert file.offset != 0, "Invalid offset"

        self.data.stream.seek(file.offset)
        tag = self.read_tag()
        assert tag == "liF#", f"Unexpected tag: {tag}"

        flag = (file.compression >> 5) & 1
        alg = file.compression & 3
        type_ = file.compression & 0xC0
        if type_ == 0x40:
            type_ = 1
            flag = 0
        elif type_ == 0x80:
            type_ = 2
        elif type_ == 0xC0:
            type_ = 3
        else:
            type_ = 0
            flag = 0

        data = self.read(file.compressed_size)
        assert len(data) == file.compressed_size, "Invalid size"
        if alg == 1:
            extracted = zlib.decompress(data, bufsize=file.size)
        elif alg == 2:
            assert data[:4] == b"(\xb5/\xfd", (
                "Expected zstd, but magic header doesn't match"
            )
            extracted = zstandard.decompress(data, file.size)
        else:
            extracted = data

        if flag == 0 and type_ == 1:
            extracted = undo_cumulative_sum(extracted)

        elif (flag == 0 and type_ in (1, 2)) or (flag == 1 and type_ == 2):
            raise NotImplementedError(
                "Stream transform functions are not implemented yet."
                f"Flag {flag}, Type {type_}"
            )

        # Check checksum
        assert len(extracted) == file.size, "File size doesn't match"
        if zlib.crc32(extracted) != file.crc32:
            warnings.warn("Invalid checksum")

        return extracted

    def close(self):
        """Close the stream."""
        self.data.stream.close()



#!/usr/bin/env python3
#
# Copyright (C) 2010 Aurelio A. Heckert, aurium (a) gmail dot com
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
#
import subprocess
import os
import sys
import tempfile

from lxml import etree

import inkex
from inkex.localization import inkex_gettext as _
from webslicer_effect import WebSlicerMixin, is_empty


class Export(WebSlicerMixin, inkex.OutputExtension):
    def add_arguments(self, pars):
        pars.add_argument("--tab")
        pars.add_argument("--dir")
        pars.add_argument("--create-dir", type=inkex.Boolean, dest="create_dir")
        pars.add_argument("--with-code", type=inkex.Boolean, dest="with_code")

    svgNS = "{http://www.w3.org/2000/svg}"

    def validate_inputs(self):
        # The user must supply a directory to export:
        if is_empty(self.options.dir):
            raise inkex.AbortExtension(
                _("You must give a directory to export the slices.")
            )
        # No directory separator at the path end:
        if self.options.dir[-1] == "/" or self.options.dir[-1] == "\\":
            self.options.dir = self.options.dir[0:-1]
        # Test if the directory exists:
        if not os.path.exists(self.options.dir):
            if self.options.create_dir:
                # Try to create it:
                try:
                    os.makedirs(self.options.dir)
                except Exception as e:
                    raise inkex.AbortExtension(
                        _("Can't create '{}': {}.".format(self.options.dir, e))
                    )
            else:
                raise inkex.AbortExtension(
                    _("Dir doesn't exist '{}'.".format(self.options.dir))
                )
        # Check whether slicer layer exists (bug #1198826)
        slicer_layer = self.get_slicer_layer()
        if slicer_layer is None:
            raise inkex.AbortExtension(_("No slicer layer found."))
        else:
            self.unique_html_id(slicer_layer)
        return None

    def get_cmd_output(self, cmd):
        try:
            pipe = subprocess.Popen(cmd)
        except FileNotFoundError:
            return 1, ""
        stdout, _ = pipe.communicate()
        sts = pipe.returncode
        if sts is None:
            sts = 0
        if stdout is not None and stdout[-1:] == "\n":
            stdout = stdout[:-1]
        return sts, stdout

    _html_ids = []

    def unique_html_id(self, el):
        for child in el.getchildren():
            if child.tag in [
                self.svgNS + "rect",
                self.svgNS + "path",
                self.svgNS + "circle",
                self.svgNS + "g",
            ]:
                conf = self.get_el_conf(child)
                if conf["html-id"] in self._html_ids:
                    inkex.errormsg(
                        'You have more than one element with "{}" html-id.'.format(
                            conf["html-id"]
                        )
                    )
                    n = 2
                    while (conf["html-id"] + "-" + str(n)) in self._html_ids:
                        n += 1
                    conf["html-id"] += "-" + str(n)
                self._html_ids.append(conf["html-id"])
                self.save_conf(conf, child)
                self.unique_html_id(child)

    def test_if_has_imagemagick(self):
        (status, output) = self.get_cmd_output("convert --version")
        self.has_magick = status == 0 and "ImageMagick" in output

    def save(self, stream):
        self.test_if_has_imagemagick()
        error = self.validate_inputs()
        if error:
            return error
        # Register the basic CSS code:
        self.reg_css("body", "text-align", "center")
        # Create the temporary SVG with invisible Slicer layer to export image pieces
        self.create_the_temporary_svg()
        # Start what we really want!
        self.export_chids_of(self.get_slicer_layer())
        # Write the HTML and CSS files if asked for:
        if self.options.with_code:
            self.make_html_file()
            self.make_css_file()
        # Delete the temporary SVG with invisible Slicer layer
        self.delete_the_temporary_svg()
        # TODO: prevent inkex to return svg code to update Inkscape

    def make_html_file(self):
        f = open(os.path.join(self.options.dir, "layout.html"), "w")
        f.write(
            "<html>\n<head>\n"
            + "  <title>Web Layout Testing</title>\n"
            + '  <style type="text/css" media="screen">\n'
            + '    @import url("style.css")\n'
            + "  </style>\n"
            + "</head>\n<body>\n"
            + self.html_code()
            + '  <p style="position:absolute; bottom:10px">\n'
            + "  This HTML code is not done to the web. <br/>\n"
            + "  The automatic HTML and CSS code are only a helper."
            + "</p>\n</body>\n</html>"
        )
        f.close()

    def make_css_file(self):
        f = open(os.path.join(self.options.dir, "style.css"), "w")
        f.write(
            "/*\n"
            + "** This CSS code is not done to the web.\n"
            + "** The automatic HTML and CSS code are only a helper.\n"
            + "*/\n"
            + self.css_code()
        )
        f.close()

    def create_the_temporary_svg(self):
        (self.tmp_svg_ref, self.tmp_svg) = tempfile.mkstemp(".svg")
        layer = self.get_slicer_layer()
        current_style = layer.style
        layer.style = "display:none"
        self.document.write(self.tmp_svg)
        layer.style = current_style

    def delete_the_temporary_svg(self):
        try:
            os.close(self.tmp_svg_ref)
            os.remove(self.tmp_svg)
        except (IOError, OSError, PermissionError):
            pass

    noid_element_count = 0

    def get_el_conf(self, el):
        desc = el.find(self.svgNS + "desc")
        conf = {}
        if desc is None:
            desc = etree.SubElement(el, "desc")
        if desc.text is None:
            desc.text = ""
        for line in desc.text.split("\n"):
            if line.find(":") > 0:
                line = line.split(":")
                conf[line[0].strip()] = line[1].strip()
        if not "html-id" in conf:
            if el == self.get_slicer_layer():
                return {"html-id": "#body#"}
            else:
                self.noid_element_count += 1
                conf["html-id"] = "element-" + str(self.noid_element_count)
                desc.text += "\nhtml-id:" + conf["html-id"]
        return conf

    def save_conf(self, conf, el):
        desc = el.find("{http://www.w3.org/2000/svg}desc")
        if desc is not None:
            conf_a = []
            for k in conf:
                conf_a.append(k + " : " + conf[k])
            desc.text = "\n".join(conf_a)

    def export_chids_of(self, parent):
        parent_id = self.get_el_conf(parent)["html-id"]
        for el in parent.getchildren():
            el_conf = self.get_el_conf(el)
            if el.tag == self.svgNS + "g":
                if self.options.with_code:
                    self.register_group_code(el, el_conf)
                else:
                    self.export_chids_of(el)
            if el.tag in [
                self.svgNS + "rect",
                self.svgNS + "path",
                self.svgNS + "circle",
            ]:
                if self.options.with_code:
                    self.register_unity_code(el, el_conf, parent_id)
                self.export_img(el, el_conf)

    def register_group_code(self, group, conf):
        self.reg_html("div", group)
        selec = "#" + conf["html-id"]
        self.reg_css(selec, "position", "absolute")
        geometry = self.get_relative_el_geometry(group)
        self.reg_css(selec, "top", str(int(geometry["y"])) + "px")
        self.reg_css(selec, "left", str(int(geometry["x"])) + "px")
        self.reg_css(selec, "width", str(int(geometry["w"])) + "px")
        self.reg_css(selec, "height", str(int(geometry["h"])) + "px")
        self.export_chids_of(group)

    def __validate_slice_conf(self, conf):
        if not "layout-disposition" in conf:
            conf["layout-disposition"] = "bg-el-norepeat"
        if not "layout-position-anchor" in conf:
            conf["layout-position-anchor"] = "mc"
        return conf

    def register_unity_code(self, el, conf, parent_id):
        conf = self.__validate_slice_conf(conf)
        css_selector = "#" + conf["html-id"]
        bg_repeat = "no-repeat"
        img_name = self.img_name(el, conf)
        if conf["layout-disposition"][0:2] == "bg":
            if conf["layout-disposition"][0:9] == "bg-parent":
                if parent_id == "#body#":
                    css_selector = "body"
                else:
                    css_selector = "#" + parent_id
                if conf["layout-disposition"] == "bg-parent-repeat":
                    bg_repeat = "repeat"
                if conf["layout-disposition"] == "bg-parent-repeat-x":
                    bg_repeat = "repeat-x"
                if conf["layout-disposition"] == "bg-parent-repeat-y":
                    bg_repeat = "repeat-y"
                lay_anchor = conf["layout-position-anchor"]
                if lay_anchor == "tl":
                    lay_anchor = "top left"
                if lay_anchor == "tc":
                    lay_anchor = "top center"
                if lay_anchor == "tr":
                    lay_anchor = "top right"
                if lay_anchor == "ml":
                    lay_anchor = "middle left"
                if lay_anchor == "mc":
                    lay_anchor = "middle center"
                if lay_anchor == "mr":
                    lay_anchor = "middle right"
                if lay_anchor == "bl":
                    lay_anchor = "bottom left"
                if lay_anchor == "bc":
                    lay_anchor = "bottom center"
                if lay_anchor == "br":
                    lay_anchor = "bottom right"
                self.reg_css(
                    css_selector,
                    "background",
                    'url("{}") {} {}'.format(img_name, bg_repeat, lay_anchor),
                )
            else:  # conf['layout-disposition'][0:9] == 'bg-el...'
                self.reg_html("div", el)
                self.reg_css(
                    css_selector,
                    "background",
                    'url("{}") {}'.format(img_name, "no-repeat"),
                )
                self.reg_css(css_selector, "position", "absolute")
                geo = self.get_relative_el_geometry(el, True)
                self.reg_css(css_selector, "top", geo["y"])
                self.reg_css(css_selector, "left", geo["x"])
                self.reg_css(css_selector, "width", geo["w"])
                self.reg_css(css_selector, "height", geo["h"])
        else:  # conf['layout-disposition'] == 'img...'
            self.reg_html("img", el)
            if conf["layout-disposition"] == "img-pos":
                self.reg_css(css_selector, "position", "absolute")
                geo = self.get_relative_el_geometry(el)
                self.reg_css(css_selector, "left", str(geo["x"]) + "px")
                self.reg_css(css_selector, "top", str(geo["y"]) + "px")
            if conf["layout-disposition"] == "img-float-left":
                self.reg_css(css_selector, "float", "right")
            if conf["layout-disposition"] == "img-float-right":
                self.reg_css(css_selector, "float", "right")

    el_geo = {}

    def register_all_els_geometry(self):
        ink_cmm = "inkscape --query-all " + self.tmp_svg
        (status, output) = self.get_cmd_output(ink_cmm)
        self.el_geo = {}
        if status == 0:
            for el in output.split("\n"):
                el = el.split(",")
                if len(el) == 5:
                    self.el_geo[el[0]] = {
                        "x": float(el[1]),
                        "y": float(el[2]),
                        "w": float(el[3]),
                        "h": float(el[4]),
                    }
        doc_w = self.svg.unittouu(self.document.getroot().get("width"))
        doc_h = self.svg.unittouu(self.document.getroot().get("height"))
        self.el_geo["webslicer-layer"] = {"x": 0, "y": 0, "w": doc_w, "h": doc_h}

    def get_relative_el_geometry(self, el, value_to_css=False):
        # This method return a dictionary with x, y, w and h keys.
        # All values are float, if value_to_css is False, otherwise
        # that is a string ended with "px". The x and y values are
        # relative to parent position.
        if not self.el_geo:
            self.register_all_els_geometry()
        parent = el.getparent()
        geometry = self.el_geo[el.attrib["id"]]
        geometry["x"] -= self.el_geo[parent.attrib["id"]]["x"]
        geometry["y"] -= self.el_geo[parent.attrib["id"]]["y"]
        if value_to_css:
            for k in geometry:
                geometry[k] = str(int(geometry[k])) + "px"
        return geometry

    def img_name(self, el, conf):
        return el.attrib["id"] + "." + conf["format"]

    def export_img(self, el, conf):
        if not self.has_magick:
            inkex.errormsg(_("You must install the ImageMagick to get JPG and GIF."))
            conf["format"] = "png"
        img_name = os.path.join(self.options.dir, self.img_name(el, conf))
        img_name_png = img_name
        if conf["format"] != "png":
            img_name_png = img_name + ".png"
        opts = ""
        if "bg-color" in conf:
            opts += ' -b "' + conf["bg-color"] + '" -y 1'
        if "dpi" in conf:
            opts += " -d " + conf["dpi"]
        if "dimension" in conf:
            dim = conf["dimension"].split("x")
            opts += " -w " + dim[0] + " -h " + dim[1]
        (status, output) = self.get_cmd_output(
            'inkscape {} -i "{}" -o "{}" "{}"'.format(
                opts, el.attrib["id"], img_name_png, self.tmp_svg
            )
        )
        if conf["format"] != "png":
            opts = ""
            if conf["format"] == "jpg":
                opts += " -quality " + str(conf["quality"])
            if conf["format"] == "gif":
                if conf["gif-type"] == "grayscale":
                    opts += " -type Grayscale"
                else:
                    opts += " -type Palette"
                if conf["palette-size"] < 256:
                    opts += " -colors " + str(conf["palette-size"])
            (status, output) = self.get_cmd_output(
                'convert "{}" {} "{}"'.format(img_name_png, opts, img_name)
            )
            if status != 0:
                inkex.errormsg("Upss... ImageMagick error: " + output)
            os.remove(img_name_png)

    _html = {}

    def reg_html(self, el_tag, el):
        parent = el.getparent()
        parent_id = self.get_el_conf(parent)["html-id"]
        if parent == self.get_slicer_layer():
            parent_id = "body"
        conf = self.get_el_conf(el)
        el_id = conf["html-id"]
        if "html-class" in conf:
            el_class = conf["html-class"]
        else:
            el_class = ""
        if not parent_id in self._html:
            self._html[parent_id] = []
        self._html[parent_id].append({"tag": el_tag, "id": el_id, "class": el_class})

    def html_code(self, parent="body", ident="  "):
        # inkex.errormsg( self._html )
        if not parent in self._html:
            return ""
        code = ""
        for el in self._html[parent]:
            child_code = self.html_code(el["id"], ident + "  ")
            tag_class = ""
            if el["class"] != "":
                tag_class = ' class="' + el["class"] + '"'
            if el["tag"] == "img":
                code += (
                    ident
                    + '<img id="'
                    + el["id"]
                    + '"'
                    + tag_class
                    + ' src="'
                    + self.img_name(el, self.get_el_conf(el))
                    + '"/>\n'
                )
            else:
                code += (
                    ident
                    + "<"
                    + el["tag"]
                    + ' id="'
                    + el["id"]
                    + '"'
                    + tag_class
                    + ">\n"
                )
                if child_code:
                    code += child_code
                else:
                    code += ident + "  Element " + el["id"] + "\n"
                code += ident + "</" + el["tag"] + '><!-- id="' + el["id"] + '" -->\n'
        return code

    _css = []

    def reg_css(self, selector, att, val):
        pos = i = -1
        for s in self._css:
            i += 1
            if s["selector"] == selector:
                pos = i
        if pos == -1:
            pos = i + 1
            self._css.append({"selector": selector, "atts": {}})
        if not att in self._css[pos]["atts"]:
            self._css[pos]["atts"][att] = []
        self._css[pos]["atts"][att].append(val)

    def css_code(self):
        code = ""
        for s in self._css:
            code += "\n" + s["selector"] + " {\n"
            for att in s["atts"]:
                val = s["atts"][att]
                if att == "background" and len(val) > 1:
                    code += "  /* the next attribute needs a CSS3 enabled browser */\n"
                code += "  " + att + ": " + (", ".join(val)) + ";\n"
            code += "}\n"
        return code


if __name__ == "__main__":
    Export().run()



#!/usr/bin/env python3
# coding=utf-8
#
# Copyright (C) 2020-2022 Jonathan Neuhauser <jonathan.neuhauser@outlook.com>
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.

"""SVG -> XAML Converter classes for structural elements 
(<g>, <svg>, <clipPath>)"""

from typing import Union, List, Optional
import inkex

from .config import GLOBAL_SETTINGS
from ..xamlobjects import (
    DrawingGroup,
    Canvas,
    GeometryGroup,
    BaseElement,
    RectangleGeometry,
    Viewbox,
    Geometry,
    KEY_ATTR,
)
from .text import TextElementParser
from .base import BaseParser
from .shapes import (
    RectangleParser,
    EllipseParser,
    LineParser,
    PolygonPolylineParser,
    PathParser,
    NonTextShapeElementParser,
)


class GroupBaseParser(BaseParser):
    """Base Parser for SVG Groups (also <svg>, <symbol> etc)"""

    def __init__(
        self, svgelement: Union[inkex.SvgDocumentElement, inkex.Group], layer=None
    ):
        super().__init__(svgelement)
        self.layer = layer
        self.child_parsers: List[BaseParser] = []
        for child in list(self.element):
            if layer is not None and (
                not isinstance(child, inkex.Group) or not child.get_id() == layer
            ):
                # ignore everything but the specified layer
                continue
            if isinstance(child, inkex.Use):
                # we unlink all clones since XAML doesn't really have a notion of clones
                # (except putting Geometries in ResourceDictionaries, which would only
                # be a very partial clone, without styles and transforms)
                child.style = child.specified_style()
                if child.href is not None:
                    child.href.style = child.href.specified_style()
                    child = child.unlink()
            if isinstance(child, inkex.TextElement):
                # We need to preprocess the text, mainly separating out tspans that do
                # not share the same line height or not simply square flowtexts
                new_elems = TextElementParser.separate_text(child)
                if new_elems:
                    last_elem = child
                    for elem in new_elems:
                        self.child_parsers.append(TextElementParser(elem))
                        last_elem.addnext(elem)
                        last_elem = elem
                    child.getparent().remove(child)
                    continue
            if isinstance(child, inkex.FlowRoot):
                # Pull the position out of the flowregion
                frs = [e for e in child if isinstance(e, inkex.FlowRegion)]
                if (
                    len(frs) == 1
                    and len(frs[0]) == 1
                    and isinstance(frs[0][0], inkex.Rectangle)
                ):
                    child.set("x", frs[0][0].left)
                    child.set("y", frs[0][0].top)
            if isinstance(child, inkex.Rectangle) and GLOBAL_SETTINGS.avalonia:
                if child.rx > 0 or child.ry > 0:
                    self.child_parsers.append(PathParser(child))
                    continue
            if child.__class__ in typemap:
                self.child_parsers.append(typemap[child.__class__](child))

    def parse_children(self, primitive: Union[DrawingGroup, Canvas, GeometryGroup]):
        """Adds the children of a groupbase to a primitive XAML object"""
        # primitive.add_property(KEY_ATTR, self.element.get_id())
        primitive.add_name(self.element.get_id())
        primitive.set_transform(self.element.transform)

        if self.element.specified_style()("display") == "none":
            primitive.set_invisible()
        if not isinstance(primitive, GeometryGroup):
            primitive.add_property(
                "ClipGeometry" if isinstance(primitive, DrawingGroup) else "Clip",
                self.get_clip(primitive),
            )
            primitive.add_property("OpacityMask", self.get_mask(primitive))
            primitive.add_property(
                "Opacity", self.element.specified_style()("opacity"), 1
            )
        blur = self.get_blur()
        # Bitmap effects are supposed to be added in Avalonia 11 or 12
        if GLOBAL_SETTINGS.wpf and blur is not None and isinstance(primitive, Canvas):
            primitive.add_property("Effect", blur)
        for parser in self.child_parsers:
            result: Optional[BaseElement] = None
            if isinstance(primitive, DrawingGroup):
                result = parser.to_xaml_objects()
            elif isinstance(primitive, Canvas):
                result = parser.to_canvas_objects()
            elif isinstance(primitive, GeometryGroup):
                if isinstance(
                    parser, NonTextShapeElementParser
                ):  # can only parse simple elements
                    result = parser.to_clippath()
            primitive.add_child(result)

    def to_xaml_objects(self) -> DrawingGroup:
        parsed = DrawingGroup()
        self.parse_children(parsed)
        return parsed

    def to_canvas_objects(self) -> Canvas:
        parsed = Canvas()
        parsed.add_property("UseLayoutRounding", "False")
        self.parse_children(parsed)
        return parsed


class SvgElementParser(GroupBaseParser):
    """Parse <svg> elements"""

    def __init__(self, svgelement: inkex.SvgDocumentElement, layer=None):
        super().__init__(svgelement, layer)

    def equivalent_transform(self):
        """Get the equivalent transform of an SVG element"""
        return inkex.Transform(
            translate=(
                self.element.unittouu(self.element.get("x", 0)),
                self.element.unittouu(self.element.get("y", 0)),
            )
        )

    def get_equivalent_clip(self) -> Optional[RectangleGeometry]:
        """Get the equivalent clip, i.e. clipped by the viewbox"""
        vbx = self.element.get_viewbox()
        clip_geometry = RectangleGeometry()
        clip_geometry.add_properties(
            {
                "Rect": ",".join(
                    map(
                        str,
                        [
                            vbx[0],
                            vbx[1],
                            self.element.viewbox_width,
                            self.element.viewbox_height,
                        ],
                    )
                )
            }
        )
        return clip_geometry

    def to_xaml_objects(self):
        self.element.transform = self.equivalent_transform() @ self.element.transform
        this = super().to_xaml_objects()

        if (
            GLOBAL_SETTINGS.layers_as_resources
            and self.element == self.element.root
            # The next two checks should already be true as a consequence of the first
            # two, but the correctness of the next operation depends on them
            and (len(this) == 1 and isinstance(this[0], DrawingGroup))
            and all(i in ["ClipGeometry"] for i in this[0].attrib.keys())
        ):
            # When exporting layers as indidual resources, there are two nested groups
            # (one for the SVG including the viewbox, which has the original layer's
            # name, and one for the layer). Iif the layer had no other relevant
            # attributes, in particular no transform, one group can be ommitted.
            if this[0].get("ClipGeometry", None) in [None, ""]:
                # Invisiblity is mocked by setting an empty ClipGeometry; we ignore
                # visibility of toplevel layers in this case.
                this = this[0]
        this.add_property("ClipGeometry", self.get_equivalent_clip())
        self.element.transform = (-self.equivalent_transform()) @ self.element.transform
        return this

    def to_canvas_objects(self):
        self.element.transform = self.equivalent_transform() @ self.element.transform
        this = super().to_canvas_objects()
        if (
            GLOBAL_SETTINGS.layers_as_resources
            and self.element == self.element.root
            # The next two checks should already be true as a consequence of the first
            # two, but the correctness of the next operation depends on them
            and len(this) == 1
            and isinstance(this[0], Canvas)
            # These attributes can be ignored safely.
            and all(
                i in ["UseLayoutRounding", "Visibility", "IsVisible"]
                for i in this[0].attrib.keys()
            )
        ):
            this = this[0]
        this.add_properties(
            {"Width": self.element.viewbox_width, "Height": self.element.viewbox_height}
        )
        this.add_property("Clip", self.get_equivalent_clip())
        result = Viewbox(nsmap=GLOBAL_SETTINGS.TOPLEVEL_DICT.nsmap)
        result.add_property("Stretch", "Uniform")
        result.add_child(this)
        self.element.transform = (-self.equivalent_transform()) @ self.element.transform
        return result


class GroupParser(GroupBaseParser):
    """Parse <g> elements"""

    def __init__(self, svgelement: Union[inkex.Group, inkex.Pattern]):
        super().__init__(svgelement)


class ClipPathParser(GroupBaseParser):
    """Parse a clippath element"""

    def __init__(self, svgelement: inkex.ClipPath):
        super().__init__(svgelement)

    def to_clippath(self) -> Optional[GeometryGroup]:
        """Convert current clippath to XAML"""
        result = GeometryGroup()
        self.parse_children(result)
        result.add_property("FillRule", "Nonzero" if GLOBAL_SETTINGS.wpf else "NonZero")
        return result

    def simplify_clip(self, clippath: GeometryGroup) -> Union[Geometry, str]:
        """Simplify the clip if possible.
        There are a few cases where we can simplify the clip.
        1. The clippath is an empty group: can be expressed as string
        2. Two nested GeometryGroups: Join the inner transforms and copy the clippath,
           if set, from the inner one one level up
        3. The top level geometrygroup has only one child
        """
        # Case 1
        if isinstance(clippath, GeometryGroup) and len(clippath) == 0:
            return ""
        # Case 2
        while len(clippath.getchildren()) == 1 and isinstance(
            clippath[0], GeometryGroup
        ):
            parent_transform = clippath.parse_transform()
            child_transform = clippath[0].parse_transform()
            clippath.set_transform(parent_transform @ child_transform)
            # clippath.add_property("FillRule", clippath[0].get("FillRule", None))
            for child in clippath[0]:
                clippath.append(child)
            clippath.remove(clippath[0])
        # Casse 3
        if len(clippath.getchildren()) == 1 and isinstance(clippath[0], Geometry):
            clippath[0].set(KEY_ATTR, clippath.get(KEY_ATTR, None))
            clippath[0].set("Transform", clippath.get("Transform", None))
            clippath = clippath[0]
        return clippath


typemap = {
    inkex.Group: GroupParser,
    inkex.Layer: GroupParser,
    inkex.SvgDocumentElement: SvgElementParser,
    inkex.PathElement: PathParser,
    inkex.Rectangle: RectangleParser,
    inkex.Circle: EllipseParser,
    inkex.Ellipse: EllipseParser,
    inkex.Polygon: PolygonPolylineParser,
    inkex.Polyline: PolygonPolylineParser,
    inkex.Line: LineParser,
    inkex.TextElement: TextElementParser,
    inkex.FlowRoot: TextElementParser,
}



#!/usr/bin/env python3
"""Custom color function per channel of selected objects"""

import ast
import operator as op
import inkex

OPS = {
    ast.Add: op.add,
    ast.Sub: op.sub,
    ast.Mult: op.mul,
    ast.Div: op.truediv,
    ast.Pow: op.pow,
    ast.BitXor: op.xor,
    ast.USub: op.neg,
}


def eval_expr(expr, namespace):
    """
    >>> eval_expr('2^6')
    4
    >>> eval_expr('2**6')
    64
    >>> eval_expr('1 + 2*3**(4^5) / (6 + -7)')
    -5.0
    """
    return _eval(ast.parse(expr, mode="eval").body, namespace)


def _eval(node, namespace):
    if isinstance(node, ast.Constant):  # <number>
        return node.value
    if isinstance(node, ast.Name):  # <variable> (must be in namespace)
        return namespace[node.id]
    if isinstance(node, ast.BinOp):  # <left> <operator> <right>
        return OPS[type(node.op)](
            _eval(node.left, namespace), _eval(node.right, namespace)
        )
    if isinstance(node, ast.UnaryOp):  # <operator> <operand> e.g., -1
        return OPS[type(node.op)](_eval(node.operand, namespace))
    raise TypeError(node)


class Custom(inkex.ColorExtension):
    """Custom colour functions per channel"""

    def add_arguments(self, pars):
        pars.add_argument("--tab")
        pars.add_argument("-r", "--r", default="r", help="red channel function")
        pars.add_argument("-g", "--g", default="g", help="green channel function")
        pars.add_argument("-b", "--b", default="b", help="blue channel function")
        pars.add_argument(
            "-s", "--scale", type=float, default=1.0, help="The input (r,g,b) range"
        )

    def modify_color(self, name, color):
        opt = self.options
        factor = 255.0 / opt.scale
        # add stuff to be accessible from within the custom color function here.
        namespace = {
            "r": float(color.red) / factor,
            "g": float(color.green) / factor,
            "b": float(color.blue) / factor,
        }

        color.red = int(eval_expr(opt.r.strip(), namespace) * factor)
        color.green = int(eval_expr(opt.g.strip(), namespace) * factor)
        color.blue = int(eval_expr(opt.b.strip(), namespace) * factor)
        return color


if __name__ == "__main__":
    Custom().run()



#!/usr/bin/env python3
"""Extension to remove the blue colour from selected shapes"""

import inkex


class RemoveBlue(inkex.ColorExtension):
    """Remove blue color from selected objects"""

    def modify_color(self, name, color):
        return inkex.Color([color.red, color.green, 0])


if __name__ == "__main__":
    RemoveBlue().run()



#!/usr/bin/env python3
"""Reduce lightness"""

import inkex


class LessLight(inkex.ColorExtension):
    """Reduce the light of the color"""

    def modify_color(self, name, color):
        color.lightness -= int(0.05 * 255)
        return color


if __name__ == "__main__":
    LessLight().run()



#!/usr/bin/env python3
# coding=utf-8
#
# Copyright (C) 2020-2022 Jonathan Neuhauser <jonathan.neuhauser@outlook.com>
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.

"""SVG -> XAML Converter classes for Text"""

from typing import List, Union, Optional
import copy
import re
import math

from lxml import etree
import inkex
from inkex.transforms import Transform
from inkex.utils import parse_percent

from .shapes import ShapeElementParser
from ..xamlobjects import (
    TextBlock,
    Span,
    Union,
    TextDecoration,
    TextDecorationCollection,
    Pen,
    BaseElement,
    DrawingGroup,
    LineBreak,
    removeNS,
    XAMLObject,
)
from .utils import convert_attributes
from .config import GLOBAL_SETTINGS


class TextElementParser(ShapeElementParser):
    """Base class for parsing text elements. This is only supported for highlevel
    representation.

    Stroke and different writing modes are not supported."""

    def __init__(self, svgelement: inkex.elements._text.TextElement):
        super().__init__(svgelement)

    def to_xaml_objects(self):
        """We'd need dx information to use <GlyphRun>, which is not available (yet)"""

    def create_obj(self):
        """Instantiate the right text type"""
        return TextBlock()

    @staticmethod
    def get_font_size(element, value=None):
        """Get the current font size"""
        if value is None:
            value = element.specified_style()("font-size")
        if not (isinstance(value, (float, int))) or value <= 0:
            return 12
        return value

    @staticmethod
    def separate_text(text: inkex.TextElement):
        """Separate the text into digestible elements"""
        # If the text already has x/y set, we don't need to do anything.
        # Also if the text has zero length, don't split and hope for the best.
        if (
            text.get("x", None) is not None
            and text.get("y", None) is not None
            or len(text) == 0
        ):
            return []
        # Create a wrapper that will be duplicated later.
        templ = copy.deepcopy(text)
        templ.text = ""
        templ.tail = ""
        for child in list(templ):
            templ.remove(child)
        y = initial_y = text.get("y", None)
        initial_x = text.get("x", None)
        result: List[inkex.TextElement] = []
        current_copy = copy.deepcopy(templ)
        dont_append = TextElementParser.get_width(text) == 0

        def finalize(result, current_copy):
            result += [current_copy]
            current_copy.set("x", initial_x)
            current_copy.set("y", initial_y)

        for tspan in text:
            style = tspan.specified_style()
            cur_y = tspan.get("y", 0)
            cur_x = tspan.get("x", 0)
            if initial_x is None or initial_y is None:
                initial_x = cur_x
                initial_y = cur_y
                append = True
            else:
                # We only support normal oriented text here.
                line_height = (
                    TextElementParser.get_line_height(tspan, style("line-height")) or 0
                )
                append = False
                # By default Inkscape only stores 6 digits
                if math.isclose(line_height, float(cur_y) - float(y), abs_tol=1e-5):
                    # This is a continuation of the previous text
                    append = True

            y = cur_y

            if not append or dont_append:
                finalize(result, current_copy)
                current_copy = copy.deepcopy(templ)
                initial_x, initial_y = cur_x, cur_y
            current_copy.add(tspan)
        if len(current_copy) > 0:
            finalize(result, current_copy)
        return result

    @staticmethod
    def convert_font_weight(value):
        """Get the equivalent font weight"""
        if value.isdigit():
            # Numeric font weight, see
            # https://learn.microsoft.com/dotnet/api/system.windows.fontweights
            return [
                "Thin",
                "ExtraLight",
                "Light",
                "Normal",
                "Medium",
                "DemiBold",
                "Bold",
                "ExtraBold",
                "Heavy",
                "UltraBlack",
            ][min((int(value) - 1) // 100, 9)]
        return {
            "lighter": "Thin",
            "normal": "Normal",
            "bold": "Bold",
            "bolder": "ExtraBlack",
        }.get(value, "Normal")

    @staticmethod
    def convert_font_family(value):
        """Get the equivalent font family"""
        # Font fallbacks are allowed, however we need to replace the generic families.
        # This is not necessarily the same result as Inkscape (which is
        # platform-dependent), but these fonts should exist on every Windows system
        # (WPF only exists on Windows). It might make sense to make this setting
        # available in the UI when this feature is available for Avalonia too.
        return ", ".join(
            [
                {
                    "sans-serif": "Arial",
                    "serif": "Times New Roman",
                    "monospace": "Courier New",
                    "fantasy": "Papyrus",
                    "cursive": "Brush Script MT",
                }
                .get(value, value)
                .replace("'", "")
                for value in map(lambda s: s.strip(), value.split(","))
            ]
        )

    @staticmethod
    def convert_font_stretch(value: str):
        """Get the equivalent font stretch"""
        dct = {
            "ultra-condensed": "UltraCondensed",
            "extra-condensed": "ExtraCondensed",
            "condensed": "Condensed",
            "semi-condensed": "SemiCondensed",
            "normal": "Medium",
            "semi-expanded": "SemiExpanded",
            "expanded": "Expanded",
            "extra-expanded": "ExtraExpanded",
            "ultra-expanded": "UltraExpanded",
        }
        if value[-1] == "%":
            # See https://learn.microsoft.com/dotnet/api/system.windows.fontstretch
            clss = max(0, int((float(value[:-1]) - 0.00001) / 12.5) - 3)
            index = 7 if clss in [7, 8] else 8 if clss > 8 else clss
            return list(dct.values())[index]
        return dct.get(value, None)

    def set_text_decoration(self, primitive: Union[Span, TextBlock]):
        """Creates a visually conforming representation of the TextDecoration.
        Does not support style: blink (deprecated). Multiple styles are supported."""
        # Get shorthand and TextDecorationLocation (decide later)
        line = self.effective_style("text-decoration-line")
        if line == "none" or line is None:
            return
        linetype = {
            "underline": "Underline",
            "overline": "Overline",
            "line-through": "Strikethrough",
        }
        style = self.effective_style("text-decoration-style")
        color = self.effective_style("text-decoration-color")
        if style == "solid" and color == "currentcolor" and line in linetype:
            # Use the predefined shorthands.
            primitive.add_property("TextDecorations", linetype[line])
            return
        collection = TextDecorationCollection()
        for line_item in line.split():
            if line_item not in linetype:
                continue
            ltype = linetype[line_item]
            # Otherwise we need to define one or more TextDecoration objects.
            result = TextDecoration()
            result.add_property("Location", ltype)
            pen = Pen()

            brush = (
                self.get_brush(primitive)
                if color == "currentcolor" or color is None
                else self.color_to_xaml_color(inkex.Color(color))
            )
            pen.add_property("Brush", brush)
            pen.add_property("Thickness", 1)  # UA choice
            result.add_property("Pen", pen)

            dashstyle_map = {
                "dashed": ("Flat", [1.5, 1.5]),
                "dotted": ("Round", [0.0, 0.75]),
            }
            if style in dashstyle_map:
                pen.add_property("DashCap", dashstyle_map[style][0])
                pen.add_property(
                    "DashStyle", self.dashstyle_from_dasharray(dashstyle_map[style][1])
                )
            collection.add_child(result)
            if style == "double":
                cpy = copy.deepcopy(result)
                cpy.add_properties(
                    {
                        "PenOffset": 0.1 if line_item == "underline" else -0.1,
                        "PenOffsetUnit": "FontRenderingEmSize",
                    }
                )
                collection.add_child(cpy)
        primitive.add_property("TextDecorations", collection)

    @staticmethod
    def get_width(element: BaseElement) -> float:
        """Figure out the width of a flowed text
        (shape-inside, inline-size, flowroot)"""
        style = element.specified_style()

        inline_size = element.to_dimensionless(style("inline-size"))
        if inline_size > 0:
            return inline_size
        # FlowRegion and shape-inside work the same way
        shape_inside = style("shape-inside")
        if shape_inside is not None and isinstance(shape_inside, inkex.Rectangle):
            if shape_inside.composed_transform() == inkex.Transform():
                return shape_inside.width

        frs = [e for e in element if isinstance(e, inkex.FlowRegion)]
        if len(frs) > 0 and len(frs[0]) == 1 and isinstance(frs[0][0], inkex.Rectangle):
            if frs[0][0].transform == inkex.Transform():
                return frs[0][0].width
        return 0

    @staticmethod
    def get_line_height(element, value=None) -> Optional[float]:
        """Compute the absolute line height"""
        if value is None:
            value = element.specified_style()("line-height")
        try:
            value = parse_percent(value) * TextElementParser.get_font_size(element)
        except ValueError:
            value = element.to_dimensionless(value)
        if value <= 0.0034 or value >= 160000:  # bounds of TextBlock.LineHeight
            return None
        return value

    def set_position_and_style(self, primitive: BaseElement):
        """Set all attributes that only exist for TextBlocks, but not Spans."""
        fontsize = self.get_font_size(self.element) or 12
        line_height = (
            self.get_line_height(self.element, self.effective_style("line-height"))
            or fontsize
        )
        top = (
            float(self.element.get("y", "0").split(" ")[0])
            - (
                3 / 2
                if isinstance(self.element, (inkex.TextElement, inkex.Tspan))
                else (1 / 2)
            )
            * fontsize
            + 1 / 2 * line_height
        )
        left = float(self.element.get("x", "0").split(" ")[0])

        primitive.add_properties({"Canvas.Top": top, "Canvas.Left": left})
        effective_transform = (
            Transform(translate=(-left, -top))
            @ self.element.transform
            @ Transform(translate=(left, top))
        )
        primitive.set_transform(effective_transform)

        if self.effective_style("display") == "none":
            primitive.set_invisible()
        primitive.add_property(
            "ClipGeometry" if isinstance(primitive, DrawingGroup) else "Clip",
            self.get_clip(primitive),
        )
        primitive.add_property("OpacityMask", self.get_mask(primitive))
        primitive.add_property("Opacity", self.effective_style("opacity"), 1)
        width = self.get_width(self.element)
        if width > 0:
            primitive.add_property("TextWrapping", "Wrap")
            primitive.add_property("Width", width)
        blur = self.get_blur()
        # Bitmap effects are supposed to be added in Avalonia 11 or 12
        if GLOBAL_SETTINGS.wpf and blur is not None:
            primitive.add_property("Effect", blur)

    def set_text_style(self, primitive: BaseElement):
        """Set all text style attributes that we know how to convert."""
        convert_attributes(
            self.effective_style,
            primitive,
            [
                [
                    "font-weight",
                    "FontWeight",
                    self.convert_font_weight,
                    "Normal",
                ],
                [
                    "font-size",
                    "FontSize",
                    lambda f: self.get_font_size(self.element, f),
                ],
                ["font-style", "FontStyle", ["normal", "italic", "oblique"], "Normal"],
                ["font-family", "FontFamily", self.convert_font_family],
                ["fill", "Foreground", self.get_brush],
                [
                    "line-height",
                    "LineHeight",
                    lambda f: self.get_line_height(self.element, f),
                ],
                # This could take unicode-bidi into account.
                [
                    "direction",
                    "FlowDirection",
                    {"ltr": "LeftToRight", "rtl": "RightToLeft"},
                    "LeftToRight",
                    "ltr",
                ],
                ["font-stretch", "FontStretch", self.convert_font_stretch, "Medium"],
                [
                    "text-align",
                    "TextAlignment",
                    {"start": "Left", "middle": "Center", "end": "Right"},
                    "Left",
                    "start",
                ],
            ],
        )
        self.set_text_decoration(primitive)

    def xmlspace(self) -> str:
        """get this element's xml:space attribute which, might be inherited"""

        def ancestors_with_self():
            yield self.element
            yield from self.element.iterancestors()

        for ancestor in ancestors_with_self():
            if "{http://www.w3.org/XML/1998/namespace}space" in ancestor.attrib:
                return ancestor.get("xml:space")
        return "default"

    def process_xmlspace(self, text):
        """Process the xml:space attribute"""
        if text is None:
            return None
        preserve = self.xmlspace() == "preserve"
        if not preserve:
            # Flatten spaces as described here:
            # https://svgwg.org/svg2-draft/text.html#LegacyXMLSpace
            return re.sub(
                r"\s+",
                " ",
                text.replace("\t", " ").replace("\r", "").replace("\n", "").strip(),
            )
        return text

    def add_spaced(self, primitive, text: str, tail: bool = False, parent=None):
        """Process text, which already has been processed according to xml:space
        For tail, the parent is needed."""
        # We need to process line breaks individually
        if text is None or primitive is None:
            return
        lines = text.split("\n")
        linebreaks = [LineBreak() for i in lines]
        setattr(primitive, "tail" if tail else "text", lines[0])
        if len(lines) == 1:
            return
        effective_parent = primitive if not tail else parent
        for line, brk in zip(lines[1:], linebreaks[1:]):
            effective_parent.add_child(brk)
            brk.tail = line

    def to_canvas_objects(self):
        if self.get_font_size(self.element) == 0:
            return None
        result = self.create_obj()
        self.add_spaced(result, self.process_xmlspace(self.element.text), False)

        self.set_position_and_style(result)
        self.set_text_style(result)
        if GLOBAL_SETTINGS.wpf:
            # This will be supported in Avalonia 11.0,
            # see https://github.com/AvaloniaUI/Avalonia/issues/2619
            for child in self.element:
                # Text element and tspans can contain other tspans
                if isinstance(child, (inkex.Tspan, inkex.FlowPara, inkex.FlowSpan)):
                    child_parsed = TspanParser(child).to_canvas_objects()
                    result.add_child(child_parsed)
                    if child.get("sodipodi:role", None) == "line":
                        result.add_child(LineBreak())
                    if isinstance(child, inkex.FlowPara):
                        result.add_child(LineBreak())
                    self.add_spaced(
                        child_parsed,
                        self.process_xmlspace(child.tail),
                        True,
                        result,
                    )
            self.simplify(result)

        return result

    def simplify(self, element: TextBlock):
        """Simplify a text recursively"""
        for child in list(element):
            if not isinstance(child, (TextBlock, Span)):
                continue  # no need to process line breaks
            # Simplify all children first.
            self.simplify(child)

            # Remove all attributes on the child that are also on the parent.
            for attr in child.attrib:
                if element.get(attr, None) == child.get(attr, None):
                    child.remove_property(attr)

            # If there are property elements, we can't integrate the element
            # into the parent.
            if any(
                "." in removeNS(super(etree.ElementBase, elem).tag)[1] for elem in child  # type: ignore
            ):
                continue

            # If the span now has no other attributes, append it to the
            # previous element's tail. If there is no previous element, append
            # it to the parent's text.
            if len(child.attrib) == 0:
                prev = child.getprevious()
                added = child.text or ""
                if prev is not None:
                    prev.tail = (prev.tail or "") + added
                else:
                    element.text = (element.text or "") + added
                # Add the remaining children (such as line breaks) one level up
                for child2 in child:
                    if isinstance(prev, XAMLObject):
                        # This unfortunately moves the tail, so we need to preserve it
                        tail = prev.tail or ""
                        prev.tail = None
                        res = prev.addnext(child2)
                        prev.tail = tail
                        prev = res
                    else:
                        prev = element.insert(0, child2)
                if prev is not None:
                    prev.tail = (prev.tail or "") + (child.tail or "")
                else:
                    element.text = (element.text or "") + (child.tail or "")

                element.remove(child)


class TspanParser(TextElementParser):
    """Convert tspans to Spans"""

    def __init__(self, svgelement: inkex.elements._text.Tspan):
        super().__init__(svgelement)

    def create_obj(self):
        return Span()

    def set_position_and_style(self, primitive: BaseElement):
        """The position (and position-related style attributes)
        are already set on the parent TextBlock, so we return immediately."""

    def set_text_style(self, primitive: BaseElement):
        super().set_text_style(primitive)
        convert_attributes(
            self.effective_style,
            primitive,
            [
                [
                    "baseline-shift",
                    "BaselineAlignment",
                    {"baseline": "Normal", "super": "Superscript", "sub": "Subscript"},
                    "Normal",
                    "baseline",
                ]
            ],
        )
        primitive.remove_property("TextAlignment")
        primitive.remove_property("LineHeight")



#!/usr/bin/env python3
# coding=utf-8
#
# Copyright (C) 2023 Jan Winkler
# based on pdflatex.py (C) by 2019 Marc Jeanmougin
#
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA
#
"""
Compile typst code directly into and insert svg into document
"""

import os

import inkex
from inkex.base import TempDirMixin
from inkex.command import ProgramRunError, call
from inkex import load_svg, ShapeElement
from inkex.units import convert_unit
from inkex.localization import inkex_gettext as _
from copy import deepcopy


class TypstFormula(TempDirMixin, inkex.GenerateExtension):
    """
    Class for compilation of typst code into svg. Ensures proper setup
    of the generated svg for insertion into Inkscape.
    """

    def add_arguments(self, pars):
        pars.add_argument(
            "--typst_code",
            type=str,
            default=r"$ sum_(k=0)^n k = 1 + ... + n = (n(n+1)) / 2 $",
        )
        pars.add_argument("--font_size", type=int, default=10)
        pars.add_argument("--page", choices=["basic"])

    def generate(self):
        """Generator for a svg group holding the formula"""
        typst_file = os.path.join(self.tempdir, "input.typ")
        svg_file = os.path.join(self.tempdir, "output.svg")

        with open(typst_file, "w") as fhl:
            self.write_typst_code(fhl)

        try:
            call("typst", "compile", typst_file, svg_file)
        except ProgramRunError as err:
            inkex.errormsg(_("An exception occurred during typst compilation: ") + "\n")
            inkex.errormsg(err.stderr.decode("utf8").replace("\r\n", "\n"))
            raise inkex.AbortExtension()

        if not os.path.isfile(svg_file):
            inkex.errormsg(_("No svg has been produced by compilation: ") + "\n")
            inkex.errormsg(_(f"Expected file name was: {svg_file}."))
            raise inkex.AbortExtension()

        with open(svg_file, "r") as fhl:
            svg: inkex.SvgDocumentElement = load_svg(fhl).getroot()
            self.expand_defs(svg)
            scale = convert_unit("1pt", "px") / self.svg.scale
            for child in svg:
                if isinstance(child, ShapeElement):
                    # Scale node correctly and also ensure that it is placed
                    # in the center of the view
                    child.transform = (
                        inkex.Transform(scale=scale)
                        @ inkex.Transform(translate=-child.bounding_box().center)
                        @ child.transform
                    )
                    yield child

    def write_typst_code(self, stream):
        """Takes a formula and wraps it in typst"""
        stream.write(
            f"""#set page(margin: (x: 0pt, y: 0pt), fill: none)
#set text({self.options.font_size}pt)
{self.options.typst_code}"""
        )
        stream.flush()

    def expand_defs(self, root):
        """Replaces references to glyph definitions by the content of the definition"""
        for child in root:
            if isinstance(child, inkex.Use):
                group = inkex.Group()
                for el in child.href:
                    group.append(deepcopy(el))
                group.transform = inkex.Transform(
                    translate=(float(child.get("x", "0")), float(child.get("y", "0")))
                )
                parent = child.getparent()
                parent.remove(child)
                parent.add(group)
                child = group  # required for recursive def
            self.expand_defs(child)


if __name__ == "__main__":
    TypstFormula().run()



# -*- coding: utf-8 -*-
#
# Copyright (c) 2020 Martin Owens <doctormo@gmail.com>
#                    Sergei Izmailov <sergei.a.izmailov@gmail.com>
#                    Thomas Holder <thomas.holder@schrodinger.com>
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
#
# pylint: disable=arguments-differ
"""
Interface for all shapes/polygons such as lines, paths, rectangles, circles etc.
"""

from __future__ import annotations

from math import cos, pi, sin
import math
from typing import Optional, Tuple, Union

from ..paths.interfaces import PathCommand
from ..paths import Arc, Curve, Move, Path, ZoneClose
from ..paths import Line as PathLine
from ..transforms import Transform, ImmutableVector2d, Vector2d
from ..bezier import pointdistance

from ._utils import addNS
from ._base import ShapeElement


class PathElementBase(ShapeElement):
    """Base element for path based shapes"""

    def get_path(self) -> Path:
        """Gets the path of the element, which can also be used in a context manager"""
        p = Path(self.get("d"))
        p.callback = self.set_path
        return p

    @classmethod
    def new(cls, path, **attrs):
        return super().new(d=Path(path), **attrs)

    def set_path(self, path):
        """Set the given data as a path as the 'd' attribute"""
        self.set("d", str(Path(path)))

    def apply_transform(self):
        """Apply the internal transformation to this node and delete"""
        if "transform" in self.attrib:
            self.path = self.path.transform(self.transform)
            self.set("transform", Transform())

    @property
    def original_path(self):
        """Returns the original path if this is a LPE, or the path if not"""
        return Path(self.get("inkscape:original-d", self.path))

    @original_path.setter
    def original_path(self, path):
        if addNS("inkscape:original-d") in self.attrib:
            self.set("inkscape:original-d", str(Path(path)))
        else:
            self.path = path


class PathElement(PathElementBase):
    """Provide a useful extension for path elements"""

    tag_name = "path"

    MAX_ARC_SUBDIVISIONS = 4

    @staticmethod
    def _arcpath(
        cx: float,
        cy: float,
        rx: float,
        ry: float,
        start: float,
        end: float,
        arctype: str,
    ) -> Optional[Path]:
        """Compute the path for an arc defined by Inkscape-specific attributes.

        For details on arguments, see :func:`arc`.

        .. versionadded:: 1.2
        """
        if abs(rx) < 1e-8 or abs(ry) < 1e-8:
            return None
        incr = end - start
        if incr < 0:
            incr += 2 * pi
        numsegs = min(1 + int(incr * 2.0 / pi), PathElement.MAX_ARC_SUBDIVISIONS)
        incr = incr / numsegs

        computed = Path()
        computed.append(Move(cos(start), sin(start)))
        for seg in range(1, numsegs + 1):
            computed.append(
                Arc(
                    1,
                    1,
                    0,
                    incr > pi,
                    1,
                    cos(start + seg * incr),
                    sin(start + seg * incr),
                )
            )
        if abs(incr * numsegs - 2 * pi) > 1e-8 and (
            arctype in ("slice", "")
        ):  # slice is default
            computed.append(PathLine(0, 0))
        if arctype != "arc":
            computed.append(ZoneClose())
        computed.transform(
            Transform().add_translate(cx, cy).add_scale(rx, ry), inplace=True
        )
        return computed.to_relative()

    @classmethod
    def arc(cls, center, rx, ry=None, arctype="", pathonly=False, **kw):  # pylint: disable=invalid-name
        """Generates a sodipodi elliptical arc (special type). Also computes the path
        that Inkscape uses under the hood.

        All data may be given as parseable strings or using numeric data types.

        Args:
            center (tuple-like): Coordinates of the star/polygon center as tuple or
                Vector2d
            rx (Union[float, str]): Radius in x direction
            ry (Union[float, str], optional): Radius in y direction. If not given,
                ry=rx. Defaults to None.
            arctype (str, optional): "arc", "chord" or "slice". Defaults to "", i.e.
                "slice".

                .. versionadded:: 1.2
                    Previously set to "arc" as fixed value
            pathonly (bool, optional): Whether to create the path without
                Inkscape-specific attributes. Defaults to False.

                .. versionadded:: 1.2
        Keyword args:
            start (Union[float, str]): start angle in radians
            end (Union[float, str]): end angle in radians
            open (str): whether the path should be open (true/false). Not used in
                Inkscape > 1.1

        Returns:
            PathElement : the created star/polygon
        """
        others = [(name, kw.pop(name, None)) for name in ("start", "end", "open")]
        elem = cls(**kw)
        elem.set("sodipodi:cx", center[0])
        elem.set("sodipodi:cy", center[1])
        elem.set("sodipodi:rx", rx)
        elem.set("sodipodi:ry", ry or rx)
        elem.set("sodipodi:type", "arc")
        if arctype != "":
            elem.set("sodipodi:arc-type", arctype)
        for name, value in others:
            if value is not None:
                elem.set("sodipodi:" + name, str(value).lower())

        path = cls._arcpath(
            float(center[0]),
            float(center[1]),
            float(rx),
            float(ry or rx),
            float(elem.get("sodipodi:start", 0)),
            float(elem.get("sodipodi:end", 2 * pi)),
            arctype,
        )
        if pathonly:
            elem = cls(**kw)
        if path is not None:
            elem.path = path
        return elem

    @classmethod
    def arc_from_3_points(
        cls,
        x: complex,
        y: complex,
        z: complex,
        arctype="slice",
    ) -> PathElement:
        """
        Create an arc through the points x, y, z.
        If those points are specified clockwise, the order is not preserved.
        This is indicated with the second return value (=clockwise)

        Returns a PathElement. May be a line if x,y,z are collinear.


        Idea: http://www.math.okstate.edu/~wrightd/INDRA/MobiusonCircles/node4.html

        .. versionadded:: 1.4
        """
        w = (z - x) / (y - x)
        if abs(w.imag) > 1e-12:
            c = -((x - y) * (w - abs(w) ** 2) / (2j * w.imag) - x)
            r = abs(c - x)

            # Now determine the arc flags by checking the angles
            deltas = [x - c, y - c, z - c]
            ang = [math.atan2(i.imag, i.real) for i in deltas]
            # Check if the angles are "in order"
            cw = int(any(ang[0 + i] < ang[-2 + i] < ang[-1 + i] for i in range(3)))
            if not cw:
                # Flip start and end angle
                ang = ang[::-1]

            return cls.arc(Vector2d(c), r, r, start=ang[0], end=ang[2], arctype=arctype)
        else:
            # Points lie on a line
            # y between x and z -> draw a line, otherwise skip
            if x.real <= y.real <= z.real or x.real >= y.real >= z.real:
                return cls.new(Path([Move(x), PathLine(z)]))
            else:
                return cls.new(Path([Move(x), Move(z)]))

    @staticmethod
    def _starpath(
        c: Tuple[float, float],
        sides: int,
        r: Tuple[float, float],  # pylint: disable=invalid-name
        arg: Tuple[float, float],
        rounded: float,
        flatsided: bool,
    ):
        """Helper method to generate the path for an Inkscape star/ polygon; randomized
        is ignored.

        For details on arguments, see :func:`star`.

        .. versionadded:: 1.2
        """

        def _star_get_xy(point, index):
            cur_arg = arg[point] + 2 * pi / sides * (index % sides)
            return Vector2d(*c) + r[point] * Vector2d(cos(cur_arg), sin(cur_arg))

        def _rot90_rel(origin, other):
            """Returns a unit length vector at 90 deg from origin to other"""
            return (
                1
                / pointdistance(other, origin)
                * Vector2d(other.y - origin.y, other.x - origin.x)
            )

        def _star_get_curvepoint(point, index, is_prev: bool):
            index = index % sides
            orig = _star_get_xy(point, index)
            previ = (index - 1 + sides) % sides
            nexti = (index + 1) % sides
            # neighbors of the current point depend on polygon or star
            prev = (
                _star_get_xy(point, previ)
                if flatsided
                else _star_get_xy(1 - point, index if point == 1 else previ)
            )
            nextp = (
                _star_get_xy(point, nexti)
                if flatsided
                else _star_get_xy(1 - point, index if point == 0 else nexti)
            )
            mid = 0.5 * (prev + nextp)
            # direction of bezier handles
            rot = _rot90_rel(orig, mid + 100000 * _rot90_rel(mid, nextp))
            ret = (
                rounded
                * rot
                * (
                    -1 * pointdistance(prev, orig)
                    if is_prev
                    else pointdistance(nextp, orig)
                )
            )
            return orig + ret

        pointy = abs(rounded) < 1e-4
        result = Path()
        result.append(Move(*_star_get_xy(0, 0)))
        for i in range(0, sides):
            # draw to point type 1 for stars
            if not flatsided:
                if pointy:
                    result.append(PathLine(*_star_get_xy(1, i)))
                else:
                    result.append(
                        Curve(
                            *_star_get_curvepoint(0, i, False),
                            *_star_get_curvepoint(1, i, True),
                            *_star_get_xy(1, i),
                        )
                    )
            # draw to point type 0 for both stars and rectangles
            if pointy and i < sides - 1:
                result.append(PathLine(*_star_get_xy(0, i + 1)))
            if not pointy:
                if not flatsided:
                    result.append(
                        Curve(
                            *_star_get_curvepoint(1, i, False),
                            *_star_get_curvepoint(0, i + 1, True),
                            *_star_get_xy(0, i + 1),
                        )
                    )
                else:
                    result.append(
                        Curve(
                            *_star_get_curvepoint(0, i, False),
                            *_star_get_curvepoint(0, i + 1, True),
                            *_star_get_xy(0, i + 1),
                        )
                    )

        result.append(ZoneClose())
        return result.to_relative()

    @classmethod
    def star(
        cls,
        center,
        radii,
        sides=5,
        rounded=0,
        args=(0, 0),
        flatsided=False,
        pathonly=False,
    ):
        """Generate a sodipodi star / polygon. Also computes the path that Inkscape uses
        under the hood. The arguments for center, radii, sides, rounded and args can be
        given as strings or as numeric data.

        .. versionadded:: 1.1

        Args:
            center (Tuple-like): Coordinates of the star/polygon center as tuple or
                Vector2d
            radii (tuple): Radii of the control points, i.e. their distances from the
                center. The control points are specified in polar coordinates. Only the
                first control point is used for polygons.
            sides (int, optional): Number of sides / tips of the polygon / star.
                Defaults to 5.
            rounded (int, optional): Controls the rounding radius of the polygon / star.
                For `rounded=0`, only straight lines are used. Defaults to 0.
            args (tuple, optional): Angle between horizontal axis and control points.
                Defaults to (0,0).

                .. versionadded:: 1.2
                    Previously fixed to (0.85, 1.3)
            flatsided (bool, optional): True for polygons, False for stars.
                Defaults to False.

                .. versionadded:: 1.2
            pathonly (bool, optional): Whether to create the path without
                Inkscape-specific attributes. Defaults to False.

                .. versionadded:: 1.2

        Returns:
            PathElement : the created star/polygon
        """
        elem = cls()
        elem.set("sodipodi:cx", center[0])
        elem.set("sodipodi:cy", center[1])
        elem.set("sodipodi:r1", radii[0])
        elem.set("sodipodi:r2", radii[1])
        elem.set("sodipodi:arg1", args[0])
        elem.set("sodipodi:arg2", args[1])
        elem.set("sodipodi:sides", max(sides, 3) if flatsided else max(sides, 2))
        elem.set("inkscape:rounded", rounded)
        elem.set("inkscape:flatsided", str(flatsided).lower())
        elem.set("sodipodi:type", "star")

        path = cls._starpath(
            (float(center[0]), float(center[1])),
            int(sides),
            (float(radii[0]), float(radii[1])),
            (float(args[0]), float(args[1])),
            float(rounded),
            flatsided,
        )
        if pathonly:
            elem = cls()
        # inkex.errormsg(path)
        if path is not None:
            elem.path = path

        return elem


class Polyline(ShapeElement):
    """Like a path, but made up of straight line segments only"""

    tag_name = "polyline"

    def get_path(self) -> Path:
        p = Path("M" + self.get("points"))
        p.callback = self.set_path
        return p

    def set_path(self, path):
        if type(path) != Path:
            path = Path("M" + str(path))
        points = [f"{x:g},{y:g}" for x, y in path.end_points]
        self.set("points", " ".join(points))

    @classmethod
    def new(cls, points=None, **attrs):
        p = super().new(**attrs)
        p.path = points
        return p


class Polygon(Polyline):
    """A closed polyline"""

    tag_name = "polygon"

    def get_path(self) -> Path:
        p = Path("M" + self.get("points") + " Z")
        p.callback = self.set_path
        return p


class Line(ShapeElement):
    """A line segment connecting two points"""

    tag_name = "line"
    x1 = property(lambda self: self.to_dimensionless(self.get("x1", 0)))
    y1 = property(lambda self: self.to_dimensionless(self.get("y1", 0)))
    x2 = property(lambda self: self.to_dimensionless(self.get("x2", 0)))
    y2 = property(lambda self: self.to_dimensionless(self.get("y2", 0)))
    get_path = lambda self: Path(f"M{self.x1},{self.y1} L{self.x2},{self.y2}")

    @classmethod
    def new(cls, start, end, **attrs):
        start = Vector2d(start)
        end = Vector2d(end)
        return super().new(x1=start.x, y1=start.y, x2=end.x, y2=end.y, **attrs)


class RectangleBase(ShapeElement):
    """Provide a useful extension for rectangle elements"""

    left = property(lambda self: self.to_dimensionless(self.get("x", "0")))
    top = property(lambda self: self.to_dimensionless(self.get("y", "0")))
    right = property(lambda self: self.left + self.width)
    bottom = property(lambda self: self.top + self.height)
    width = property(lambda self: self.to_dimensionless(self.get("width", "0")))
    height = property(lambda self: self.to_dimensionless(self.get("height", "0")))
    rx = property(
        lambda self: self.to_dimensionless(self.get("rx", self.get("ry", 0.0)))
    )
    ry = property(
        lambda self: self.to_dimensionless(self.get("ry", self.get("rx", 0.0)))
    )  # pylint: disable=invalid-name

    def get_path(self) -> Path:
        """Calculate the path as the box around the rect"""
        if self.rx or self.ry:
            # pylint: disable=invalid-name
            rx = min(self.rx if self.rx > 0 else self.ry, self.width / 2)
            ry = min(self.ry if self.ry > 0 else self.rx, self.height / 2)
            cpts = [self.left + rx, self.right - rx, self.top + ry, self.bottom - ry]
            return Path(
                f"M {cpts[0]},{self.top}"
                f"L {cpts[1]},{self.top}    "
                f"A {rx},{ry} 0 0 1 {self.right},{cpts[2]}"
                f"L {self.right},{cpts[3]}  "
                f"A {rx},{ry} 0 0 1 {cpts[1]},{self.bottom}"
                f"L {cpts[0]},{self.bottom} "
                f"A {rx},{ry} 0 0 1 {self.left},{cpts[3]}"
                f"L {self.left},{cpts[2]}   "
                f"A {rx},{ry} 0 0 1 {cpts[0]},{self.top} z"
            )

        return Path(
            f"M {self.left},{self.top} h{self.width}v{self.height}h{-self.width} z"
        )


class Rectangle(RectangleBase):
    """Provide a useful extension for rectangle elements"""

    tag_name = "rect"

    @classmethod
    def new(cls, left, top, width, height, **attrs):
        return super().new(x=left, y=top, width=width, height=height, **attrs)


class EllipseBase(ShapeElement):
    """Absorbs common part of Circle and Ellipse classes"""

    def get_path(self) -> Path:
        """Calculate the arc path of this circle"""
        rx, ry = self.rxry()
        cx, y = self.center.x, self.center.y - ry
        return Path(
            (
                "M {cx},{y} "
                "a {rx},{ry} 0 1 0 {rx}, {ry} "
                "a {rx},{ry} 0 0 0 -{rx}, -{ry} z"
            ).format(cx=cx, y=y, rx=rx, ry=ry)
        )

    @property
    def center(self):
        """Return center of circle/ellipse"""
        return ImmutableVector2d(
            self.to_dimensionless(self.get("cx", "0")),
            self.to_dimensionless(self.get("cy", "0")),
        )

    @center.setter
    def center(self, value):
        value = Vector2d(value)
        self.set("cx", value.x)
        self.set("cy", value.y)

    def rxry(self):
        # type: () -> Vector2d
        """Helper function"""
        raise NotImplementedError()

    @classmethod
    def new(cls, center, radius, **attrs):
        circle = super().new(**attrs)
        circle.center = center
        circle.radius = radius
        return circle


class Circle(EllipseBase):
    """Provide a useful extension for circle elements"""

    tag_name = "circle"

    @property
    def radius(self) -> float:
        """Return radius of circle"""
        return self.to_dimensionless(self.get("r", "0"))

    @radius.setter
    def radius(self, value):
        self.set("r", self.to_dimensionless(value))

    def rxry(self):
        r = self.radius
        return Vector2d(r, r)


class Ellipse(EllipseBase):
    """Provide a similar extension to the Circle interface for ellipses"""

    tag_name = "ellipse"

    @property
    def radius(self) -> ImmutableVector2d:
        """Return radii of ellipse"""
        return ImmutableVector2d(
            self.to_dimensionless(self.get("rx", "0")),
            self.to_dimensionless(self.get("ry", "0")),
        )

    @radius.setter
    def radius(self, value):
        value = Vector2d(value)
        self.set("rx", str(value.x))
        self.set("ry", str(value.y))

    def rxry(self):
        return self.radius



# coding=utf-8
#
# Copyright (C) 2018 Martin Owens <doctormo@gmail.com>
# Copyright (C) 2023 Jonathan Neuhauser <jonathan.neuhauser@outlook.com>
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
#
"""Arc path commands"""

from __future__ import annotations
from math import atan2, pi, sqrt, sin, cos, tan, acos, radians, degrees
from cmath import exp
from typing import overload, Tuple, List, Union, TYPE_CHECKING

import numpy as np

from ..transforms import Transform, Vector2d, ComplexLike

from .interfaces import (
    AbsolutePathCommand,
    RelativePathCommand,
    LengthSettings,
    ILengthSettings,
    BezierArcComputationMixin,
)

if TYPE_CHECKING:
    from .curves import Curve


class Arc(BezierArcComputationMixin, AbsolutePathCommand):
    """Special Arc segment"""

    letter = "A"

    nargs = 7

    radius: complex
    """Radius of the Arc"""
    x_axis_rotation: float

    large_arc: bool
    sweep: bool

    endpoint: complex
    """Endpoint (absolute) of the Arc"""

    @property
    def rx(self) -> float:
        """x radius of the Arc"""
        return self.radius.real

    @property
    def ry(self) -> float:
        """y radius of the Arc"""
        return self.radius.imag

    @property
    def x(self) -> float:
        """x coordinate of the (absolute) endpoint of the Arc"""
        return self.endpoint.real

    @property
    def y(self) -> float:
        """x coordinate of the (relative) endpoint of the Arc"""
        return self.endpoint.imag

    @property
    def args(self):
        return (
            self.rx,
            self.ry,
            self.x_axis_rotation,
            self.large_arc,
            self.sweep,
            self.x,
            self.y,
        )

    @property
    def cargs(self):
        """Set of arguments in complex form"""
        return (
            self.radius,
            self.x_axis_rotation,
            self.large_arc,
            self.sweep,
            self.endpoint,
        )

    @overload
    def __init__(
        self,
        radius: ComplexLike,
        x_axis_rotation: float,
        large_arc: bool | int,
        sweep: bool | int,
        endpoint: ComplexLike,
    ) -> None: ...

    @overload
    def __init__(
        self,
        rx: float,
        ry: float,
        x_axis_rotation: float,
        large_arc: bool | int,
        sweep: bool | int,
        x: float,
        y: float,
    ) -> None: ...  # pylint: disable=too-many-arguments

    def __init__(self, *args):
        if len(args) == 5:
            (
                self.radius,
                self.x_axis_rotation,
                self.large_arc,
                self.sweep,
                self.endpoint,
            ) = args
            self.radius = complex(self.radius)
            self.endpoint = complex(self.endpoint)
        elif len(args) == 7:
            self.radius = args[0] + args[1] * 1j
            self.x_axis_rotation, self.large_arc, self.sweep = args[2:5]
            self.endpoint = args[5] + args[6] * 1j

    def parametrize(self, prev):
        """Return the parameterization of the arc

        See http://www.w3.org/TR/SVG/implnote.html#ArcImplementationNotes

        Returns:
            A tuple (radius, phi, rot_matrix, center, theta1, deltatheta)

        .. versionadded:: 1.4
        """
        # my notation roughly follows theirs
        phi = radians(self.x_axis_rotation)
        rot_matrix = exp(1j * phi)

        radius = self.radius
        rx = self.radius.real
        ry = self.radius.imag
        rx_sqd = rx * rx
        ry_sqd = ry * ry

        # Transform z-> z' = x' + 1j*y'
        # = self.rot_matrix**(-1)*(z - (end+start)/2)
        # coordinates.  This translates the ellipse so that the midpoint
        # between self.end and self.start lies on the origin and rotates
        # the ellipse so that the its axes align with the xy-coordinate axes.
        # Note:  This sends self.end to -self.start
        zp1 = (1 / rot_matrix) * (prev - self.cend_point(0j, prev)) / 2
        x1p, y1p = zp1.real, zp1.imag
        x1p_sqd = x1p * x1p
        y1p_sqd = y1p * y1p

        # Correct out of range radii
        radius_check = (x1p_sqd / rx_sqd) + (y1p_sqd / ry_sqd)
        if radius_check > 1:
            rx *= sqrt(radius_check)
            ry *= sqrt(radius_check)
            radius = rx + 1j * ry
            rx_sqd = rx * rx
            ry_sqd = ry * ry

        # Compute c'=(c_x', c_y'), the center of the ellipse in (x', y') coords
        # Noting that, in our new coord system, (x_2', y_2') = (-x_1', -x_2')
        # and our ellipse is cut out by of the plane by the algebraic equation
        # (x'-c_x')**2 / r_x**2 + (y'-c_y')**2 / r_y**2 = 1,
        # we can find c' by solving the system of two quadratics given by
        # plugging our transformed endpoints (x_1', y_1') and (x_2', y_2')
        tmp = rx_sqd * y1p_sqd + ry_sqd * x1p_sqd
        radicand = (rx_sqd * ry_sqd - tmp) / tmp
        radical = 0 if np.isclose(radicand, 0) else sqrt(radicand)

        if self.large_arc == self.sweep:
            cp = -radical * (rx * y1p / ry - 1j * ry * x1p / rx)
        else:
            cp = radical * (rx * y1p / ry - 1j * ry * x1p / rx)

        # The center in (x,y) coordinates is easy to find knowing c'
        center = exp(1j * phi) * cp + (prev + self.cend_point(0j, prev)) / 2

        # Now we do a second transformation, from (x', y') to (u_x, u_y)
        # coordinates, which is a translation moving the center of the
        # ellipse to the origin and a dilation stretching the ellipse to be
        # the unit circle
        u1 = (x1p - cp.real) / rx + 1j * (y1p - cp.imag) / ry  # transformed start
        u2 = (-x1p - cp.real) / rx + 1j * (-y1p - cp.imag) / ry  # transformed end

        # clip in case of floating point error
        u1 = np.clip(u1.real, -1, 1) + 1j * np.clip(u1.imag, -1, 1)
        u2 = np.clip(u2.real, -1, 1) + 1j * np.clip(u2.imag, -1, 1)

        # Now compute theta and delta (we'll define them as we go)
        # delta is the angular distance of the arc (w.r.t the circle)
        # theta is the angle between the positive x'-axis and the start point
        # on the circle
        if u1.imag > 0:
            theta1 = degrees(acos(u1.real))
        elif u1.imag < 0:
            theta1 = -degrees(acos(u1.real))
        else:
            if u1.real > 0:  # start is on pos u_x axis
                theta1 = 0
            else:  # start is on neg u_x axis
                # Note: This behavior disagrees with behavior documented in
                # http://www.w3.org/TR/SVG/implnote.html#ArcImplementationNotes
                # where theta is set to 0 in this case.
                theta1 = 180

        det_uv = u1.real * u2.imag - u1.imag * u2.real

        acosand = u1.real * u2.real + u1.imag * u2.imag
        acosand = np.clip(acosand.real, -1, 1) + np.clip(acosand.imag, -1, 1)

        if det_uv > 0:
            deltatheta = degrees(acos(acosand))
        elif det_uv < 0:
            deltatheta = -degrees(acos(acosand))
        else:
            if u1.real * u2.real + u1.imag * u2.imag > 0:
                # u1 == u2
                deltatheta = 0
            else:
                # u1 == -u2
                # Note: This behavior disagrees with behavior documented in
                # http://www.w3.org/TR/SVG/implnote.html#ArcImplementationNotes
                # where deltatheta is set to 0 in this case.
                deltatheta = 180

        if not self.sweep and deltatheta >= 0:
            deltatheta -= 360
        elif self.large_arc and deltatheta <= 0:
            deltatheta += 360

        return radius, phi, rot_matrix, center, theta1, deltatheta

    def update_bounding_box(self, first, last_two_points, bbox):
        prev = last_two_points[-1]
        for seg in self.to_curves(prev=prev):
            seg.update_bounding_box(first, [None, prev], bbox)
            prev = seg.cend_point(first, prev)

    def ccontrol_points(
        self, first: complex, prev: complex, prev_prev: complex
    ) -> Tuple[complex, ...]:
        return (self.endpoint,)

    def ccurve_points(
        self, first: complex, prev: complex, prev_prev: complex
    ) -> Tuple[complex, ...]:
        return NotImplemented

    def to_curves(self, prev: ComplexLike, prev_prev: ComplexLike = 0j) -> List[Curve]:
        """Convert this arc into bezier curves"""
        # TODO Refactor out CubicSuperPath
        from .path import CubicSuperPath

        path = CubicSuperPath(
            [arc_to_path(Vector2d.c2t(complex(prev)), self.args)]
        ).to_path(curves_only=True)
        # Ignore the first move command from to_path()
        return list(path)[1:]

    def transform(self, transform: Transform) -> Arc:
        # pylint: disable=invalid-name, too-many-locals
        newend = transform.capply_to_point(self.endpoint)

        T: Transform = transform
        if self.x_axis_rotation != 0:
            T = T @ Transform(rotate=self.x_axis_rotation)
        a, c, b, d, _, _ = list(T.to_hexad())
        # T = | a b |
        #     | c d |

        detT = a * d - b * c
        detT2 = detT**2

        rx = float(self.rx)
        ry = float(self.ry)

        def get_degen():
            return Arc(
                self.radius,
                self.x_axis_rotation,
                self.large_arc,
                self.sweep,
                newend,
            )

        if rx == 0.0 or ry == 0.0 or detT2 == 0.0:
            # degenerate arc
            # transform only last point
            return get_degen()
        A = (d**2 / rx**2 + c**2 / ry**2) / detT2
        B = -(d * b / rx**2 + c * a / ry**2) / detT2
        D = (b**2 / rx**2 + a**2 / ry**2) / detT2

        theta = atan2(-2 * B, D - A) / 2
        theta_deg = theta * 180.0 / pi
        DA = D - A
        l2 = 4 * B**2 + DA**2

        if l2 == 0:
            delta = 0.0
        else:
            delta = 0.5 * (-(DA**2) - 4 * B**2) / sqrt(l2)

        half = (A + D) / 2

        try:
            rx_ = 1.0 / sqrt(half + delta)
            ry_ = 1.0 / sqrt(half - delta)

            if detT > 0:
                sweep = self.sweep
            else:
                sweep = not self.sweep > 0
            return Arc(rx_ + 1j * ry_, theta_deg, self.large_arc, sweep, newend)
        except ZeroDivisionError:
            return get_degen()

    def to_relative(self, prev: ComplexLike) -> RelativePathCommand:
        return arc(
            self.radius,
            self.x_axis_rotation,
            self.large_arc,
            self.sweep,
            self.endpoint - prev,
        )

    def cend_point(self, first: complex, prev: complex) -> complex:
        return self.endpoint

    def reverse(self, first: ComplexLike, prev: ComplexLike) -> Arc:
        return Arc(
            self.radius, self.x_axis_rotation, self.large_arc, not self.sweep, prev
        )

    def _cpoint(
        self, first: complex, prev: complex, prev_control: complex, t: float
    ) -> complex:
        # TODO surely this can be expressed in a shorter way using complex geometry?
        radius, _, rot_matrix, center, theta1, deltatheta = self.parametrize(prev)
        angle = (theta1 + t * deltatheta) * pi / 180
        cosphi = rot_matrix.real
        sinphi = rot_matrix.imag
        rx = radius.real
        ry = radius.imag

        x = rx * cosphi * cos(angle) - ry * sinphi * sin(angle) + center.real
        y = rx * sinphi * cos(angle) + ry * cosphi * sin(angle) + center.imag
        return x + y * 1j

    def _cderivative(
        self, first: complex, prev: complex, prev_control: complex, t: float, n: int = 1
    ) -> complex:
        """returns the nth derivative of the segment at t."""
        radius, phi, _, _, theta1, deltatheta = self.parametrize(prev)
        angle = radians(theta1 + t * deltatheta)
        rx = radius.real
        ry = radius.imag
        k = (deltatheta * pi / 180) ** n  # ((d/dt)angle)**n

        if n % 4 == 0 and n > 0:
            return (
                rx * cos(phi) * cos(angle)
                - ry * sin(phi) * sin(angle)
                + 1j * (rx * sin(phi) * cos(angle) + ry * cos(phi) * sin(angle))
            )
        elif n % 4 == 1:
            return k * (
                -rx * cos(phi) * sin(angle)
                - ry * sin(phi) * cos(angle)
                + 1j * (-rx * sin(phi) * sin(angle) + ry * cos(phi) * cos(angle))
            )
        elif n % 4 == 2:
            return k * (
                -rx * cos(phi) * cos(angle)
                + ry * sin(phi) * sin(angle)
                + 1j * (-rx * sin(phi) * cos(angle) - ry * cos(phi) * sin(angle))
            )
        elif n % 4 == 3:
            return k * (
                rx * cos(phi) * sin(angle)
                + ry * sin(phi) * cos(angle)
                + 1j * (rx * sin(phi) * sin(angle) - ry * cos(phi) * cos(angle))
            )
        else:
            raise ValueError("n should be a positive integer.")

    def _split(
        self, first: complex, prev: complex, prev_control: complex, t: float
    ) -> Tuple[Arc, Arc]:
        """returns two segments, whose union is this segment and which join
        at self.point(t)."""
        radius, _, _, _, _, deltatheta = self.parametrize(prev)

        def crop(t0, t1):
            return Arc(
                radius,
                self.x_axis_rotation,
                not abs(deltatheta * (t1 - t0)) <= 180,
                self.sweep,
                self.cpoint(0j, prev, 0j, t1),
            )

        return crop(0, t), crop(t, 1)

    def _ilength(
        self,
        first: complex,
        prev: complex,
        prev_control: complex,
        length: float,
        settings: ILengthSettings = ILengthSettings(),
    ):
        # ilength calls self.parametrize very often, so we cache the result
        param = self.parametrize
        params = param(prev)

        def cached(prev):  # pylint: disable=unused-argument
            return params

        self.parametrize = cached  # type: ignore
        try:
            return super()._ilength(first, prev, prev_control, length, settings)
        finally:
            self.parametrize = param  # type: ignore


class arc(RelativePathCommand, Arc):  # pylint: disable=invalid-name
    """Relative Arc line segment"""

    letter = "a"

    nargs = 7

    endpoint: complex
    """Endpoint (relative) of the arc"""

    @property
    def rx(self) -> float:
        """x radius of the arc"""
        return self.radius.real

    @property
    def ry(self) -> float:
        """y radius of the arc"""
        return self.radius.imag

    @property
    def dx(self) -> float:
        """x coordinate of the (relative) endpoint of the arc"""
        return self.endpoint.real

    @property
    def dy(self) -> float:
        """x coordinate of the (relative) endpoint of the arc"""
        return self.endpoint.imag

    @property
    def args(self):
        return (
            self.rx,
            self.ry,
            self.x_axis_rotation,
            self.large_arc,
            self.sweep,
            self.dx,
            self.dy,
        )

    @overload
    def __init__(
        self,
        radius: ComplexLike,
        x_axis_rotation: float,
        large_arc: bool,
        sweep: bool,
        endpoint: ComplexLike,
    ) -> None: ...

    @overload
    def __init__(
        self,
        rx: float,
        ry: float,
        x_axis_rotation: float,
        large_arc: bool,
        sweep: bool,
        dx: float,
        dy: float,
    ) -> None: ...  # pylint: disable=too-many-arguments

    def __init__(self, *args):
        if len(args) == 5:
            (
                self.radius,
                self.x_axis_rotation,
                self.large_arc,
                self.sweep,
                self.endpoint,
            ) = args
            self.radius = complex(self.radius)
            self.endpoint = complex(self.endpoint)
        elif len(args) == 7:
            self.radius = args[0] + args[1] * 1j
            self.x_axis_rotation, self.large_arc, self.sweep = args[2:5]
            self.endpoint = args[5] + args[6] * 1j

    def to_absolute(self, prev: ComplexLike) -> Arc:
        return Arc(
            self.radius,
            self.x_axis_rotation,
            self.large_arc,
            self.sweep,
            self.endpoint + prev,
        )

    def cend_point(self, first: complex, prev: complex) -> complex:
        return self.endpoint + prev

    def ccontrol_points(
        self, first: complex, prev: complex, prev_prev: complex
    ) -> Tuple[complex, ...]:
        return (self.endpoint + prev,)

    def ccurve_points(
        self, first: complex, prev: complex, prev_prev: complex
    ) -> Tuple[complex, ...]:
        return NotImplemented

    def reverse(self, first: ComplexLike, prev: ComplexLike) -> arc:
        return arc(
            self.radius,
            self.x_axis_rotation,
            self.large_arc,
            not self.sweep,
            -self.endpoint,
        )

    def to_curves(self, prev: ComplexLike, prev_prev: ComplexLike = 0j) -> List[Curve]:
        return self.to_absolute(prev).to_curves(prev, prev_prev)


def arc_to_path(point, params):
    """Approximates an arc with cubic bezier segments.

    Args:
        point:  Starting point (absolute coords)
        params: Arcs parameters as per
              https://www.w3.org/TR/SVG/paths.html#PathDataEllipticalArcCommands

    Returns:
        A list of triplets of points
        `[control_point_before, node, control_point_after]`
        (first and last returned triplets are `[p1, p1, *]` and `[*, p2, p2]`)
    """

    # pylint: disable=invalid-name, too-many-locals
    A = point[:]
    rx, ry, teta, longflag, sweepflag, x2, y2 = params[:]
    teta = teta * pi / 180.0
    B = [x2, y2]
    # Degenerate ellipse
    if rx == 0 or ry == 0 or A == B:
        return [[A[:], A[:], A[:]], [B[:], B[:], B[:]]]

    # turn coordinates so that the ellipse morph into a *unit circle* (not 0-centered)
    mat = matprod((rotmat(teta), [[1.0 / rx, 0.0], [0.0, 1.0 / ry]], rotmat(-teta)))
    applymat(mat, A)
    applymat(mat, B)

    k = [-(B[1] - A[1]), B[0] - A[0]]
    d = k[0] * k[0] + k[1] * k[1]
    k[0] /= sqrt(d)
    k[1] /= sqrt(d)
    d = sqrt(max(0, 1 - d / 4.0))
    # k is the unit normal to AB vector, pointing to center O
    # d is distance from center to AB segment (distance from O to the midpoint of AB)
    # for the last line, remember this is a unit circle, and kd vector is ortogonal to
    # AB (Pythagorean thm)

    if longflag == sweepflag:
        # top-right ellipse in SVG example
        # https://www.w3.org/TR/SVG/images/paths/arcs02.svg
        d *= -1

    O = [(B[0] + A[0]) / 2.0 + d * k[0], (B[1] + A[1]) / 2.0 + d * k[1]]
    OA = [A[0] - O[0], A[1] - O[1]]
    OB = [B[0] - O[0], B[1] - O[1]]
    start = acos(OA[0] / norm(OA))
    if OA[1] < 0:
        start *= -1
    end = acos(OB[0] / norm(OB))
    if OB[1] < 0:
        end *= -1
    # start and end are the angles from center of the circle to A and to B respectively

    if sweepflag and start > end:
        end += 2 * pi
    if (not sweepflag) and start < end:
        end -= 2 * pi

    NbSectors = int(abs(start - end) * 2 / pi) + 1
    dTeta = (end - start) / NbSectors
    v = 4 * tan(dTeta / 4.0) / 3.0
    # I would use v = tan(dTeta/2)*4*(sqrt(2)-1)/3 ?
    p = []
    for i in range(0, NbSectors + 1, 1):
        angle = start + i * dTeta
        v1 = [
            O[0] + cos(angle) - (-v) * sin(angle),
            O[1] + sin(angle) + (-v) * cos(angle),
        ]
        pt = [O[0] + cos(angle), O[1] + sin(angle)]
        v2 = [O[0] + cos(angle) - v * sin(angle), O[1] + sin(angle) + v * cos(angle)]
        p.append([v1, pt, v2])
    p[0][0] = p[0][1][:]
    p[-1][2] = p[-1][1][:]

    # go back to the original coordinate system
    mat = matprod((rotmat(teta), [[rx, 0], [0, ry]], rotmat(-teta)))
    for pts in p:
        applymat(mat, pts[0])
        applymat(mat, pts[1])
        applymat(mat, pts[2])
    return p


def matprod(mlist):
    """Get the product of the mat"""
    prod = mlist[0]
    for mat in mlist[1:]:
        a00 = prod[0][0] * mat[0][0] + prod[0][1] * mat[1][0]
        a01 = prod[0][0] * mat[0][1] + prod[0][1] * mat[1][1]
        a10 = prod[1][0] * mat[0][0] + prod[1][1] * mat[1][0]
        a11 = prod[1][0] * mat[0][1] + prod[1][1] * mat[1][1]
        prod = [[a00, a01], [a10, a11]]
    return prod


def rotmat(teta):
    """Rotate the mat"""
    return [[cos(teta), -sin(teta)], [sin(teta), cos(teta)]]


def applymat(mat, point):
    """Apply the given mat"""
    x = mat[0][0] * point[0] + mat[0][1] * point[1]
    y = mat[1][0] * point[0] + mat[1][1] * point[1]
    point[0] = x
    point[1] = y


def norm(point):
    """Normalise"""
    return sqrt(point[0] * point[0] + point[1] * point[1])



#!/usr/bin/env python3
# coding=utf-8
"""
Test elements extra logic from svg xml lxml custom classes.
"""

import os
import sys
from importlib import resources

from lxml import etree

from ..utils import PY3
from ..inx import InxFile

INTERNAL_ARGS = ("help", "output", "id", "selected-nodes")
ARG_TYPES = {
    "Boolean": "bool",
    "Color": "color",
    "str": "string",
    "int": "int",
    "float": "float",
}


class InxMixin:
    """Tools for Testing INX files, use as a mixin class:

    class MyTests(InxMixin, TestCase):
        def test_inx_file(self):
            self.assertInxIsGood("some_inx_file.inx")
    """

    def assertInxIsGood(self, inx_file):  # pylint: disable=invalid-name
        """Test the inx file for consistancy and correctness"""
        self.assertTrue(PY3, "INX files can only be tested in python3")

        inx = InxFile(inx_file)
        if "help" in inx.ident or inx.script.get("interpreter", None) != "python":
            return
        cls = inx.extension_class
        # Check class can be matched in python file
        self.assertTrue(cls, f"Can not find class for {inx.filename}")
        # Check name is reasonable for the class
        if not cls.multi_inx:
            self.assertEqual(
                cls.__name__,
                inx.slug,
                f"Name of extension class {cls.__module__}.{cls.__name__} "
                f"is different from ident {inx.slug}",
            )
            self.assertParams(inx, cls)

    def assertParams(self, inx, cls):  # pylint: disable=invalid-name
        """Confirm the params in the inx match the python script

        .. versionchanged:: 1.2
            Also checks that the default values are identical
        """
        params = {param.name: self.parse_param(param) for param in inx.params}
        args = dict(self.introspect_arg_parser(cls().arg_parser))
        mismatch_a = list(set(params) ^ set(args) & set(params))
        mismatch_b = list(set(args) ^ set(params) & set(args))
        self.assertFalse(
            mismatch_a, f"{inx.filename}: Inx params missing from arg parser"
        )
        self.assertFalse(
            mismatch_b, f"{inx.filename}: Script args missing from inx xml"
        )

        for param in args:
            if params[param]["type"] and args[param]["type"]:
                self.assertEqual(
                    params[param]["type"],
                    args[param]["type"],
                    f"Type is not the same for {inx.filename}:param:{param}",
                )
            inxdefault = params[param]["default"]
            argsdefault = args[param]["default"]
            if inxdefault and argsdefault:
                # for booleans, the inx is lowercase and the param is uppercase
                if params[param]["type"] == "bool":
                    argsdefault = str(argsdefault).lower()
                elif params[param]["type"] not in ["string", None, "color"] or args[
                    param
                ]["type"] in ["int", "float"]:
                    # try to parse the inx value to compare numbers to numbers
                    inxdefault = float(inxdefault)
                if args[param]["type"] == "color" or callable(args[param]["default"]):
                    # skip color, method types
                    continue
                self.assertEqual(
                    argsdefault,
                    inxdefault,
                    f"Default value is not the same for {inx.filename}:param:{param}",
                )
            inxchoices = params[param]["choices"]
            argschoices = args[param]["choices"]
            if argschoices is not None and len(argschoices) > 0:
                assert set(inxchoices).issubset(argschoices), (
                    f"params don't match: inx={inxchoices}, py={argschoices}"
                )

    def introspect_arg_parser(self, arg_parser):
        """Pull apart the arg parser to find out what we have in it"""
        for action in arg_parser._optionals._actions:  # pylint: disable=protected-access
            for opt in action.option_strings:
                # Ignore params internal to inkscape (thus not in the inx)
                if opt.startswith("--") and opt[2:] not in INTERNAL_ARGS:
                    yield (opt[2:], self.introspect_action(action))

    @staticmethod
    def introspect_action(action):
        """Pull apart a single action to get at the juicy insides"""
        return {
            "type": ARG_TYPES.get((action.type or str).__name__, "string"),
            "default": action.default,
            "choices": action.choices,
            "help": action.help,
        }

    @staticmethod
    def parse_param(param):
        """Pull apart the param element in the inx file"""
        if param.param_type in ("optiongroup", "notebook"):
            options = param.options
            return {
                "type": None,
                "choices": options,
                "default": options and options[0] or None,
            }
        param_type = param.param_type
        if param.param_type in ("path",):
            param_type = "string"
        return {
            "type": param_type,
            "default": param.text,
            "choices": None,
        }

    def assertInxSchemaValid(self, inx_file):  # pylint: disable=invalid-name
        """Validate inx file schema."""
        self.assertTrue(INX_SCHEMAS, "no schema files found")
        with open(inx_file, "rb") as fp:
            inx_doc = etree.parse(fp)

        for schema_name, schema in INX_SCHEMAS.items():
            with self.subTest(schema_file=schema_name):
                schema.assert_(inx_doc)


def _load_inx_schemas():
    _SCHEMA_CLASSES = {
        ".rng": etree.RelaxNG,
        ".schema": etree.Schematron,  # "pre-ISO-Schematron"
    }

    if sys.version_info > (3, 9):

        def _contents(pkg):
            return [path.name for path in resources.files(pkg).iterdir()]
    else:
        _contents = resources.contents

    for name in _contents(__package__):
        _, ext = os.path.splitext(name)
        schema_class = _SCHEMA_CLASSES.get(ext)
        if schema_class is None:
            continue

        if sys.version_info > (3, 9):

            def _open_binary(pkg, res):
                return resources.files(pkg).joinpath(res).open("rb")
        else:
            _open_binary = resources.open_binary

        with _open_binary(__package__, name) as fp:
            schema_doc = etree.parse(fp)
        yield name, schema_class(schema_doc)


INX_SCHEMAS = dict(_load_inx_schemas())
